##
# .protocol.element3
##
"""
PQ version 3.0 elements.
"""
import sys
import os
import pprint
from struct import unpack, Struct
from .message_types import message_types
from ..python.structlib import ushort_pack, ushort_unpack, ulong_pack, ulong_unpack

try:
	from ..port.optimized import parse_tuple_message, pack_tuple_data
except ImportError:
	def pack_tuple_data(atts,
		none = None,
		ulong_pack = ulong_pack,
		blen = bytes.__len__
	):
		return b''.join([
			b'\xff\xff\xff\xff'
			if x is none
			else (ulong_pack(blen(x)) + x)
			for x in atts
		])

try:
	from ..port.optimized import cat_messages
except ImportError:
	from ..python.structlib import lH_pack, long_pack
	# Special case tuple()'s
	def _pack_tuple(t,
		blen = bytes.__len__,
		tlen = tuple.__len__,
		pack_head = lH_pack,
		ulong_pack = ulong_pack,
		ptd = pack_tuple_data,
	):
		# NOTE: duplicated from above
		r = b''.join([
			b'\xff\xff\xff\xff'
			if x is None
			else (ulong_pack(blen(x)) + x)
			for x in t
		])
		return pack_head((blen(r) + 6, tlen(t))) + r

	def cat_messages(messages,
		lpack = long_pack,
		blen = bytes.__len__,
		tuple = tuple,
		pack_tuple = _pack_tuple
	):
		return b''.join([
			(x.bytes() if x.__class__ is not bytes else (
				b'd' + lpack(blen(x) + 4) + x
			)) if x.__class__ is not tuple else (
				b'D' + pack_tuple(x)
			) for x in messages
		])
	del _pack_tuple, lH_pack, long_pack

StringFormat = b'\x00\x00'
BinaryFormat = b'\x00\x01'

class Message(object):
	bytes_struct = Struct("!cL")
	__slots__ = ()
	def __repr__(self):
		return '%s.%s(%s)' %(
			type(self).__module__,
			type(self).__name__,
			', '.join([repr(getattr(self, x)) for x in self.__slots__])
		)

	def __eq__(self, ob):
		return isinstance(ob, type(self)) and self.type == ob.type and \
		not False in (
			getattr(self, x) == getattr(ob, x)
			for x in self.__slots__
		)

	def bytes(self):
		data = self.serialize()
		return self.bytes_struct.pack(self.type, len(data) + 4) + data

	@classmethod
	def parse(typ, data):
		return typ(data)

class StringMessage(Message):
	"""
	A message based on a single string component.
	"""
	type = b''
	__slots__ = ('data',)

	def __repr__(self):
		return '%s.%s(%s)' %(
			type(self).__module__,
			type(self).__name__,
			repr(self.data),
		)

	def __getitem__(self, i):
		return self.data.__getitem__(i)

	def __init__(self, data):
		self.data = data

	def serialize(self):
		return bytes(self.data) + b'\x00'

	@classmethod
	def parse(typ, data):
		if not data.endswith(b'\x00'):
			raise ValueError("string message not NUL-terminated")
		return typ(data[:-1])

class TupleMessage(tuple, Message):
	"""
	A message who's data is based on a tuple structure.
	"""
	type = b''
	__slots__ = ()

	def __repr__(self):
		return '%s.%s(%s)' %(
			type(self).__module__,
			type(self).__name__,
			tuple.__repr__(self)
		)

class Void(Message):
	"""
	An absolutely empty message. When serialized, it always yields an empty string.
	"""
	type = b''
	__slots__ = ()

	def bytes(self):
		return b''

	def serialize(self):
		return b''

	def __new__(typ, *args, **kw):
		return VoidMessage
VoidMessage = Message.__new__(Void)

def dict_message_repr(self):
	return '%s.%s(**%s)' %(
		type(self).__module__,
		type(self).__name__,
		pprint.pformat(dict(self))
	)

class WireMessage(Message):
	def __init__(self, typ_data):
		self.type = message_types[typ_data[0][0]]
		self.data = typ_data[1]

	def serialize(self):
		return self[1]

	@classmethod
	def parse(typ, data):
		if ulong_unpack(data[1:5]) != len(data) - 1:
			raise ValueError(
				"invalid wire message where data is %d bytes and " \
				"internal size stamp is %d bytes" %(
					len(data), ulong_unpack(data[1:5]) + 1
				)
			)
		return typ((data[0:1], data[5:]))

class EmptyMessage(Message):
	"""
	An abstract message that is always empty.
	"""
	__slots__ = ()
	type = b''

	def __new__(typ):
		return typ.SingleInstance

	def serialize(self):
		return b''

	@classmethod
	def parse(typ, data):
		if data != b'':
			raise ValueError("empty message(%r) had data" %(typ.type,))
		return typ.SingleInstance

class Notify(Message):
	"""
	Asynchronous notification message.
	"""
	type = message_types[b'A'[0]]
	__slots__ = ('pid', 'channel', 'payload',)

	def __init__(self, pid, channel, payload = b''):
		self.pid = pid
		self.channel = channel
		self.payload = payload

	def serialize(self):
		return ulong_pack(self.pid) + \
			self.channel + b'\x00' + \
			self.payload + b'\x00'

	@classmethod
	def parse(typ, data):
		pid = ulong_unpack(data)
		channel, payload, _ = data[4:].split(b'\x00', 2)
		return typ(pid, channel, payload)

class ShowOption(Message):
	"""
	GUC variable information from backend
	"""
	type = message_types[b'S'[0]]
	__slots__ = ('name', 'value')

	def __init__(self, name, value):
		self.name = name
		self.value = value

	def serialize(self):
		return self.name + b'\x00' + self.value + b'\x00'

	@classmethod
	def parse(typ, data):
		return typ(*(data.split(b'\x00', 2)[0:2]))

class Complete(StringMessage):
	"""
	Command completion message.
	"""
	type = message_types[b'C'[0]]
	__slots__ = ()

	@classmethod
	def parse(typ, data):
		return typ(data.rstrip(b'\x00'))

	def extract_count(self):
		"""
		Extract the last set of digits as an integer.
		"""
		# Find the last sequence of digits.
		# If there are no fields consisting only of digits, there is no count.
		for x in reversed(self.data.split()):
			if x.isdigit():
				return int(x)
		return None

	def extract_command(self):
		"""
		Strip all the *surrounding* digits and spaces from the command tag,
		and return that string.
		"""
		return self.data.strip(b'\c\n\t 0123456789') or None

class Null(EmptyMessage):
	"""
	Null command.
	"""
	type = message_types[b'I'[0]]
	__slots__ = ()
NullMessage = Message.__new__(Null)
Null.SingleInstance = NullMessage

class NoData(EmptyMessage):
	"""
	Null command.
	"""
	type = message_types[b'n'[0]]
	__slots__ = ()
NoDataMessage = Message.__new__(NoData)
NoData.SingleInstance = NoDataMessage

class ParseComplete(EmptyMessage):
	"""
	Parse reaction.
	"""
	type = message_types[b'1'[0]]
	__slots__ = ()
ParseCompleteMessage = Message.__new__(ParseComplete)
ParseComplete.SingleInstance = ParseCompleteMessage

class BindComplete(EmptyMessage):
	"""
	Bind reaction.
	"""
	type = message_types[b'2'[0]]
	__slots__ = ()
BindCompleteMessage = Message.__new__(BindComplete)
BindComplete.SingleInstance = BindCompleteMessage

class CloseComplete(EmptyMessage):
	"""
	Close statement or Portal.
	"""
	type = message_types[b'3'[0]]
	__slots__ = ()
CloseCompleteMessage = Message.__new__(CloseComplete)
CloseComplete.SingleInstance = CloseCompleteMessage

class Suspension(EmptyMessage):
	"""
	Portal was suspended, more tuples for reading.
	"""
	type = message_types[b's'[0]]
	__slots__ = ()
SuspensionMessage = Message.__new__(Suspension)
Suspension.SingleInstance = SuspensionMessage

class Ready(Message):
	"""
	Ready for new query message.
	"""
	type = message_types[b'Z'[0]]
	possible_states = (
		message_types[b'I'[0]],
		message_types[b'E'[0]],
		message_types[b'T'[0]],
	)
	__slots__ = ('xact_state',)

	def __init__(self, data):
		if data not in self.possible_states:
			raise ValueError("invalid state for Ready message: " + repr(data))
		self.xact_state = data

	def serialize(self):
		return self.xact_state

class Notice(Message, dict):
	"""
	Notification message.

	Used by PQ to emit INFO, NOTICE, and WARNING messages among other
	severities.
	"""
	type = message_types[b'N'[0]]
	__slots__ = ()
	__repr__ = dict_message_repr

	def serialize(self):
		return b'\x00'.join([
			k + v for k, v in self.items()
			if k and v is not None
		]) + b'\x00'

	@classmethod
	def parse(typ, data, msgtypes = message_types):
		return typ([
			(msgtypes[x[0]], x[1:])
			# "if x" reduce empty fields
			for x in data.split(b'\x00') if x
		])

class ClientNotice(Notice):
	__slots__ = ()

	def serialize(self):
		raise RuntimeError("cannot serialize ClientNotice")

	@classmethod
	def parse(self):
		raise RuntimeError("cannot parse ClientNotice")

class Error(Notice):
	"""
	Error information message.
	"""
	type = message_types[b'E'[0]]
	__slots__ = ()

class ClientError(Error):
	__slots__ = ()

	def serialize(self):
		raise RuntimeError("cannot serialize ClientError")

	@classmethod
	def parse(self):
		raise RuntimeError("cannot serialize ClientError")

class FunctionResult(Message):
	"""
	Function result value.
	"""
	type = message_types[b'V'[0]]
	__slots__ = ('result',)

	def __init__(self, datum):
		self.result = datum

	def serialize(self):
		return self.result is None and b'\xff\xff\xff\xff' or \
			ulong_pack(len(self.result)) + self.result

	@classmethod
	def parse(typ, data):
		if data == b'\xff\xff\xff\xff':
			return typ(None)
		size = ulong_unpack(data[0:4])
		data = data[4:]
		if size != len(data):
			raise ValueError(
				"data length(%d) is not equal to the specified message size(%d)" %(
					len(data), size
				)
			)
		return typ(data)

class AttributeTypes(TupleMessage):
	"""
	Tuple attribute types.
	"""
	type = message_types[b't'[0]]
	__slots__ = ()

	def serialize(self):
		return ushort_pack(len(self)) + b''.join([ulong_pack(x) for x in self])

	@classmethod
	def parse(typ, data):
		ac = ushort_unpack(data[0:2])
		args = data[2:]
		if len(args) != ac * 4:
			raise ValueError("invalid argument type data size")
		return typ(unpack('!%dL'%(ac,), args))

class TupleDescriptor(TupleMessage):
	"""
	Tuple structure description.
	"""
	type = message_types[b'T'[0]]
	struct = Struct("!LhLhlh")
	__slots__ = ()

	def keys(self):
		return [x[0] for x in self]

	def serialize(self):
		return ushort_pack(len(self)) + b''.join([
			x[0] + b'\x00' + self.struct.pack(*x[1:])
			for x in self
		])

	@classmethod
	def parse(typ, data):
		ac = ushort_unpack(data[0:2])
		atts = []
		data = data[2:]
		ca = 0
		while ca < ac:
			# End Of Attribute Name
			eoan = data.index(b'\x00')
			name = data[0:eoan]
			data = data[eoan+1:]
			# name, relationId, columnNumber, typeId, typlen, typmod, format
			atts.append((name,) + typ.struct.unpack(data[0:18]))
			data = data[18:]
			ca += 1
		return typ(atts)

class Tuple(TupleMessage):
	"""
	Tuple Data.
	"""
	type = message_types[b'D'[0]]
	__slots__ = ()

	def serialize(self):
		return ushort_pack(len(self)) + pack_tuple_data(self)

	@classmethod
	def parse(typ, data,
		T = tuple, ulong_unpack = ulong_unpack,
		len = len
	):
		natts = ushort_unpack(data[0:2])
		atts = []
		offset = 2
		add = atts.append

		while natts > 0:
			alo = offset
			offset += 4
			size = data[alo:offset]
			if size == b'\xff\xff\xff\xff':
				att = None
			else:
				al = ulong_unpack(size)
				ao = offset
				offset = ao + al
				att = data[ao:offset]
			add(att)
			natts -= 1
		return T(atts)
	try:
		parse = parse_tuple_message
	except NameError:
		# This is an override when port.optimized is available.
		pass

class KillInformation(Message):
	"""
	Backend cancellation information.
	"""
	type = message_types[b'K'[0]]
	struct = Struct("!LL")
	__slots__ = ('pid', 'key')

	def __init__(self, pid, key):
		self.pid = pid
		self.key = key

	def serialize(self):
		return self.struct.pack(self.pid, self.key)

	@classmethod
	def parse(typ, data):
		return typ(*typ.struct.unpack(data))

class CancelRequest(KillInformation):
	"""
	Abort the query in the specified backend.
	"""
	type = b''
	from .version import CancelRequestCode as version
	packed_version = version.bytes()
	__slots__ = ('pid', 'key')

	def serialize(self):
		return self.packed_version + self.struct.pack(
			self.pid, self.key
		)

	def bytes(self):
		data = self.serialize()
		return ulong_pack(len(data) + 4) + self.serialize()

	@classmethod
	def parse(typ, data):
		if data[0:4] != typ.packed_version:
			raise ValueError("invalid cancel query code")
		return typ(*typ.struct.unpack(data[4:]))

class NegotiateSSL(Message):
	"""
	Discover backend's SSL support.
	"""
	type = b''
	from .version import NegotiateSSLCode as version
	packed_version = version.bytes()
	__slots__ = ()

	def __new__(typ):
		return NegotiateSSLMessage

	def bytes(self):
		data = self.serialize()
		return ulong_pack(len(data) + 4) + data

	def serialize(self):
		return self.packed_version

	@classmethod
	def parse(typ, data):
		if data != typ.packed_version:
			raise ValueError("invalid SSL Negotiation code")
		return NegotiateSSLMessage
NegotiateSSLMessage = Message.__new__(NegotiateSSL)

class Startup(Message, dict):
	"""
	Initiate a connection using the given keywords.
	"""
	type = b''
	from py_opengauss.protocol.version import V3_51 as version
	packed_version = version.bytes()
	__slots__ = ()
	__repr__ = dict_message_repr

	def serialize(self):
		return self.packed_version + b''.join([
			k + b'\x00' + v + b'\x00'
			for k, v in self.items()
			if v is not None
		]) + b'\x00'

	def bytes(self):
		data = self.serialize()
		return ulong_pack(len(data) + 4) + data

	@classmethod
	def parse(typ, data):
		if data[0:4] != typ.packed_version:
			raise ValueError("invalid version code {1}".format(repr(data[0:4])))
		kw = dict()
		key = None
		for value in data[4:].split(b'\x00')[:-2]:
			if key is None:
				key = value
				continue
			kw[key] = value
			key = None
		return typ(kw)

AuthRequest_OK = 0
AuthRequest_Cleartext = 3
AuthRequest_Password = AuthRequest_Cleartext
AuthRequest_Crypt = 4
AuthRequest_MD5 = 5
AuthRequest_SHA256 = 10 # implementation for opengauss

# Unsupported by pg_protocol.
AuthRequest_KRB4 = 1
AuthRequest_KRB5 = 2
AuthRequest_SCMC = 6
AuthRequest_SSPI = 9
AuthRequest_GSS = 7
AuthRequest_GSSContinue = 8

AuthNameMap = {
	AuthRequest_Password : 'Cleartext',
	AuthRequest_Crypt : 'Crypt',
	AuthRequest_MD5 : 'MD5',

	AuthRequest_KRB4 : 'Kerberos4',
	AuthRequest_KRB5 : 'Kerberos5',
	AuthRequest_SCMC : 'SCM Credential',
	AuthRequest_SSPI : 'SSPI',
	AuthRequest_GSS : 'GSS',
	AuthRequest_GSSContinue : 'GSSContinue',
	AuthRequest_SHA256: 'SHA256',
}

class Authentication(Message):
	"""
	Authentication(request, salt)
	"""
	type = message_types[b'R'[0]]
	__slots__ = ('request', 'salt')

	def __init__(self, request, salt):
		self.request = request
		self.salt = salt

	def serialize(self):
		return ulong_pack(self.request) + self.salt

	@classmethod
	def parse(typ, data):
		return typ(ulong_unpack(data[0:4]), data[4:])

class Password(StringMessage):
	"""
	Password supplement.
	"""
	type = message_types[b'p'[0]]
	__slots__ = ('data',)

class Disconnect(EmptyMessage):
	"""
	Connection closed message.
	"""
	type = message_types[b'X'[0]]
	__slots__ = ()
DisconnectMessage = Message.__new__(Disconnect)
Disconnect.SingleInstance = DisconnectMessage

class Flush(EmptyMessage):
	"""
	Flush message.
	"""
	type = message_types[b'H'[0]]
	__slots__ = ()
FlushMessage = Message.__new__(Flush)
Flush.SingleInstance = FlushMessage

class Synchronize(EmptyMessage):
	"""
	Synchronize.
	"""
	type = message_types[b'S'[0]]
	__slots__ = ()
SynchronizeMessage = Message.__new__(Synchronize)
Synchronize.SingleInstance = SynchronizeMessage

class Query(StringMessage):
	"""
	Execute the query with the given arguments.
	"""
	type = message_types[b'Q'[0]]
	__slots__ = ('data',)

class Parse(Message):
	"""
	Parse a query with the specified argument types.
	"""
	type = message_types[b'P'[0]]
	__slots__ = ('name', 'statement', 'argtypes')

	def __init__(self, name, statement, argtypes):
		self.name = name
		self.statement = statement
		self.argtypes = argtypes

	@classmethod
	def parse(typ, data):
		name, statement, args = data.split(b'\x00', 2)
		ac = ushort_unpack(args[0:2])
		args = args[2:]
		if len(args) != ac * 4:
			raise ValueError("invalid argument type data")
		at = unpack('!%dL'%(ac,), args)
		return typ(name, statement, at)

	def serialize(self):
		ac = ushort_pack(len(self.argtypes))
		return self.name + b'\x00' + self.statement + b'\x00' + ac + b''.join([
			ulong_pack(x) for x in self.argtypes
		])

class Bind(Message):
	"""
	Bind a parsed statement with the given arguments to a Portal

	Bind(
		name,      # Portal/Cursor identifier
		statement, # Prepared Statement name/identifier
		aformats,  # Argument formats; Sequence of BinaryFormat or StringFormat.
		arguments, # Argument data; Sequence of None or argument data(str).
		rformats,  # Result formats; Sequence of BinaryFormat or StringFormat.
	)
	"""
	type = message_types[b'B'[0]]
	__slots__ = ('name', 'statement', 'aformats', 'arguments', 'rformats')

	def __init__(self, name, statement, aformats, arguments, rformats):
		self.name = name
		self.statement = statement
		self.aformats = aformats
		self.arguments = arguments
		self.rformats = rformats

	def serialize(self, len = len):
		args = self.arguments
		ac = ushort_pack(len(args))
		ad = pack_tuple_data(tuple(args))
		return \
			self.name + b'\x00' + self.statement + b'\x00' + \
			ac + b''.join(self.aformats) + ac + ad + \
			ushort_pack(len(self.rformats)) + b''.join(self.rformats)

	@classmethod
	def parse(typ, message_data):
		name, statement, data = message_data.split(b'\x00', 2)
		ac = ushort_unpack(data[:2])
		offset = 2 + (2 * ac)
		aformats = unpack(("2s" * ac), data[2:offset])

		natts = ushort_unpack(data[offset:offset+2])
		args = list()
		offset += 2

		while natts > 0:
			alo = offset
			offset += 4
			size = data[alo:offset]
			if size == b'\xff\xff\xff\xff':
				att = None
			else:
				al = ulong_unpack(size)
				ao = offset
				offset = ao + al
				att = data[ao:offset]
			args.append(att)
			natts -= 1

		rfc = ushort_unpack(data[offset:offset+2])
		ao = offset + 2
		offset = ao + (2 * rfc)
		rformats = unpack(("2s" * rfc), data[ao:offset])

		return typ(name, statement, aformats, args, rformats)

class Execute(Message):
	"""
	Fetch results from the specified Portal.
	"""
	type = message_types[b'E'[0]]
	__slots__ = ('name', 'max')

	def __init__(self, name, max = 0):
		self.name = name
		self.max = max

	def serialize(self):
		return self.name + b'\x00' + ulong_pack(self.max)

	@classmethod
	def parse(typ, data):
		name, max = data.split(b'\x00', 1)
		return typ(name, ulong_unpack(max))

class Describe(StringMessage):
	"""
	Request a description of a Portal or Prepared Statement.
	"""
	type = message_types[b'D'[0]]
	__slots__ = ('data',)

	def serialize(self):
		return self.subtype + self.data + b'\x00'

	@classmethod
	def parse(typ, data):
		if data[0:1] != typ.subtype:
			raise ValueError(
				"invalid Describe message subtype, %r; expected %r" %(
					typ.subtype, data[0:1]
				)
			)
		return super().parse(data[1:])

class DescribeStatement(Describe):
	subtype = message_types[b'S'[0]]
	__slots__ = ('data',)

class DescribePortal(Describe):
	subtype = message_types[b'P'[0]]
	__slots__ = ('data',)

class Close(StringMessage):
	"""
	Generic Close
	"""
	type = message_types[b'C'[0]]
	__slots__ = ()

	def serialize(self):
		return self.subtype + self.data + b'\x00'

	@classmethod
	def parse(typ, data):
		if data[0:1] != typ.subtype:
			raise ValueError(
				"invalid Close message subtype, %r; expected %r" %(
					typ.subtype, data[0:1]
				)
			)
		return super().parse(data[1:])

class CloseStatement(Close):
	"""
	Close the specified Statement
	"""
	subtype = message_types[b'S'[0]]
	__slots__ = ()

class ClosePortal(Close):
	"""
	Close the specified Portal
	"""
	subtype = message_types[b'P'[0]]
	__slots__ = ()

class Function(Message):
	"""
	Execute the specified function with the given arguments
	"""
	type = message_types[b'F'[0]]
	__slots__ = ('oid', 'aformats', 'arguments', 'rformat')

	def __init__(self, oid, aformats, args, rformat):
		self.oid = oid
		self.aformats = aformats
		self.arguments = args
		self.rformat = rformat

	def serialize(self):
		ac = ushort_pack(len(self.arguments))
		return ulong_pack(self.oid) + \
			ac + b''.join(self.aformats) + \
			ac + pack_tuple_data(tuple(self.arguments)) + self.rformat

	@classmethod
	def parse(typ, data):
		oid = ulong_unpack(data[0:4])

		ac = ushort_unpack(data[4:6])
		offset = 6 + (2 * ac)
		aformats = unpack(("2s" * ac), data[6:offset])

		natts = ushort_unpack(data[offset:offset+2])
		args = list()
		offset += 2

		while natts > 0:
			alo = offset
			offset += 4
			size = data[alo:offset]
			if size == b'\xff\xff\xff\xff':
				att = None
			else:
				al = ulong_unpack(size)
				ao = offset
				offset = ao + al
				att = data[ao:offset]
			args.append(att)
			natts -= 1

		return typ(oid, aformats, args, data[offset:])

class CopyBegin(Message):
	type = None
	struct = Struct("!BH")
	__slots__ = ('format', 'formats')

	def __init__(self, format, formats):
		self.format = format
		self.formats = formats

	def serialize(self):
		return self.struct.pack(self.format, len(self.formats)) + b''.join([
			ushort_pack(x) for x in self.formats
		])

	@classmethod
	def parse(typ, data):
		format, natts = typ.struct.unpack(data[:3])
		formats_str = data[3:]
		if len(formats_str) != natts * 2:
			raise ValueError("number of formats and data do not match up")
		return typ(format, [
			ushort_unpack(formats_str[x:x+2]) for x in range(0, natts * 2, 2)
		])

class CopyToBegin(CopyBegin):
	"""
	Begin copying to.
	"""
	type = message_types[b'H'[0]]
	__slots__ = ('format', 'formats')

class CopyFromBegin(CopyBegin):
	"""
	Begin copying from.
	"""
	type = message_types[b'G'[0]]
	__slots__ = ('format', 'formats')

class CopyData(Message):
	type = message_types[b'd'[0]]
	__slots__ = ('data',)

	def __init__(self, data):
		self.data = bytes(data)

	def serialize(self):
		return self.data

	@classmethod
	def parse(typ, data):
		return typ(data)

class CopyFail(StringMessage):
	type = message_types[b'f'[0]]
	__slots__ = ('data',)

class CopyDone(EmptyMessage):
	type = message_types[b'c'[0]]
	__slots__ = ('data',)
CopyDoneMessage = Message.__new__(CopyDone)
CopyDone.SingleInstance = CopyDoneMessage



##
# .copyman - COPY manager
##
"""
Manage complex COPY operations; one-to-many COPY streaming.

Primarily this module houses the `CopyManager` class, and the `transfer`
function for a high-level interface to using the `CopyManager`.
"""
import sys
from abc import abstractmethod, abstractproperty
from collections import Iterator
from .python.element import Element, ElementSet
from .python.structlib import ulong_unpack, ulong_pack
from .protocol.buffer import pq_message_stream
from .protocol.element3 import CopyData, CopyDone, Complete, cat_messages
from .protocol.xact3 import Complete as xactComplete

#: 10KB buffer for COPY messages by default.
default_buffer_size = 1024 * 10

class Fault(Exception):
	pass

class ProducerFault(Fault):
	"""
	Exception raised when the Producer caused an exception.

	Normally, Producer faults are fatal.
	"""
	def __init__(self, manager):
		self.manager = manager

	def __str__(self):
		return "producer raised exception"

class ReceiverFault(Fault):
	"""
	Exception raised when Receivers cause an exception.

	Faults should be trapped if recovery from an exception is
	possible, or if the failed receiver is optional to the succes of the
	operation.

	The 'manager' attribute is the CopyManager that raised the fault.

	The 'faults' attribute is a dictionary mapping the receiver to the exception
	instance raised.
	"""
	def __init__(self, manager, faults):
		self.manager = manager
		self.faults = faults

	def __str__(self):
		return "{0} faults occurred".format(len(self.faults))

class CopyFail(Exception):
	"""
	Exception thrown by the CopyManager when the COPY operation failed.

	The 'manager' attribute the CopyManager that raised the CopyFail.

	The 'reason' attribute is a string indicating why it failed.

	The 'receiver_faults' attribute is a mapping of receivers to exceptions that were
	raised on exit.

	The 'producer_fault' attribute specifies if the producer raise an exception
	on exit.
	"""
	def __init__(self, manager, reason = None,
		receiver_faults = None,
		producer_fault = None,
	):
		self.manager = manager
		self.reason = reason
		self.receiver_faults = receiver_faults or {}
		self.producer_fault = producer_fault

	def __str__(self):
		return self.reason or 'copy aborted'

# The identifier for PQv3 copy data.
PROTOCOL_PQv3 = "PQv3"
# The identifier for iterables of copy data sequences.
# iter([[row1, row2], [row3, row4]])
PROTOCOL_CHUNKS = "CHUNKS"
# The protocol identifier for NULL producers and receivers.
PROTOCOL_NULL = None

class ChunkProtocol(object):
	__slots__ = ('buffer',)
	def __init__(self):
		self.buffer = pq_message_stream()

	def __call__(self, data):
		self.buffer.write(bytes(data))
		return [x[1] for x in self.buffer.read()]

# Null protocol mapping.
def EmptyView(arg):
	return memoryview(b'')
def EmptyList(arg):
	return []
def ReturnNone(arg):
	return None
# Zero-Transformation
def NoTransformation(arg):
	return arg

# Copy protocols being at the Python level; *not* wire/serialization format.
copy_protocol_mappings = {
	# PQv3 -> Chunks
	(PROTOCOL_PQv3, PROTOCOL_CHUNKS) : ChunkProtocol,
	# Chunks -> PQv3
	(PROTOCOL_CHUNKS, PROTOCOL_PQv3) : lambda: cat_messages,
	# Null Producers and Receivers
	(PROTOCOL_NULL, PROTOCOL_PQv3) : lambda: EmptyView,
	(PROTOCOL_NULL, PROTOCOL_CHUNKS) : lambda: EmptyList,
	(PROTOCOL_PQv3, PROTOCOL_NULL) : lambda: ReturnNone,
	(PROTOCOL_CHUNKS, PROTOCOL_NULL) : lambda: ReturnNone,
	# Zero Transformations
	(PROTOCOL_NULL, PROTOCOL_NULL) : lambda: NoTransformation,
	(PROTOCOL_CHUNKS, PROTOCOL_CHUNKS) : lambda: NoTransformation,
	(PROTOCOL_PQv3, PROTOCOL_PQv3) : lambda: NoTransformation,
}

# Used to manage the conversions of COPY data.
# Notably, chunks -> PQv3 or PQv3 -> chunks.
class CopyTransformer(object):
	__slots__ = ('current', 'transformers', 'get')
	def __init__(self, source_protocol, target_protocols):
		self.current = {}
		self.transformers = {
			x : copy_protocol_mappings[(source_protocol, x)]()
			for x in set(target_protocols)
		}
		self.get = self.current.__getitem__

	def __call__(self, data):
		for protocol, transformer in self.transformers.items():
			self.current[protocol] = transformer(data)

##
# This is the object that does the magic.
# It tracks the state of the wire.
# It ends when non-COPY data is found.
class WireState(object):
	"""
	Manages the state of the wire.

	This class manages three possible positions:

	 1. Between wire messages
	 2. Inside message header
	 3. Inside message (with complete header)

	The wire state will become unusable when the configured condition is True.
	"""
	__slots__ = ('remaining_bytes', 'size_fragment', 'final_view', 'condition',)

	def update(self, view, getlen = ulong_unpack, len = len):
		"""
		Given the state of the COPY and new data, advance the position on the
		COPY stream.
		"""
		# Only usable until the terminating condition.
		if self.final_view is not None:
			raise RuntimeError("wire state encountered exceptional condition")

		nmessages = 0

		# State carried over from prior run.
		remaining_bytes = self.remaining_bytes
		size_fragment = self.size_fragment

		# Terminating condition.
		CONDITION = self.condition

		# Is it a continuation of a message header?
		if remaining_bytes == -1:
			##
			# Inside message header; after message type.
			# Continue adding to the 'size_fragment'
			# until there are four bytes to unpack.
			##
			o = len(size_fragment)
			size_fragment += bytes(view[:4-o])
			if len(size_fragment) == 4:
				# The size fragment is completed; only part
				# of the fragment remains to be consumed.
				remaining_bytes = getlen(size_fragment) - o
				size_fragment = b''
			else:
				assert len(size_fragment) < 4
				# size_fragment got updated..

		if remaining_bytes >= 0:
			vlen = len(view)
			while True:
				if remaining_bytes:
					##
					# Inside message body. Message length has been unpacked.
					##
					view = view[remaining_bytes:]
					# How much is remaining now?
					rb = remaining_bytes - vlen
					if rb <= 0:
						# Finished it.
						vlen = -rb
						remaining_bytes = 0
						nmessages += 1
					else:
						vlen = 0
						remaining_bytes = rb
				##
				# In between protocol messages.
				##
				if not view:
					# no more data to analyze
					break
				# There is at least one byte in the view.
				if CONDITION(view[0]):
					# State is dead now.
					# User needs to handle unexpected message, then continue.
					self.final_view = view
					assert remaining_bytes == 0
					break
				if vlen < 5:
					# Header continuation.
					remaining_bytes = -1
					view = view[1:]
					size_fragment += bytes(view)
					# Not enough left for the header of the next message?
					break
				# Update remaining_bytes to include the header, and start over.
				remaining_bytes = getlen(view[1:5]) + 1

		# Update the state for the next update.
		self.remaining_bytes, self.size_fragment = (
			remaining_bytes, size_fragment,
		)
		# Emit the number of messages "consumed" this round.
		return nmessages

	def __init__(self, condition = (CopyData.type[0].__ne__ if isinstance(memoryview(b'f')[0], int) else CopyData.type.__ne__)):
		self.remaining_bytes = 0
		self.size_fragment = b''
		self.final_view = None
		self.condition = condition

class Fitting(Element):
	_e_label = 'FITTING'

	def _e_metas(self):
		yield None, '[' + self.state + ']'

	@abstractproperty
	def protocol(self):
		"""
		The COPY data format produced or consumed.
		"""

	# Used to setup the Receiver/Producer
	def __enter__(self):
		pass

	# Used to tear down the Receiver/Producer
	def __exit__(self, typ, val, tb):
		pass

class Producer(Fitting, Iterator):
	_e_label = 'PRODUCER'

	def _e_metas(self):
		for x in super()._e_metas():
			yield x
		yield 'data', str(self.total_bytes / (1024**2)) + 'MB'
		yield 'messages', self.total_messages
		yield 'average size', (self.total_bytes / self.total_messages)

	def __init__(self):
		self.total_messages = 0
		self.total_bytes = 0

	@abstractmethod
	def realign(self):
		"""
		Method implemented by producers that emit COPY data that is not
		guaranteed to be aligned.

		This is only necessary in failure cases where receivers still need more
		data to complete the message.
		"""

	@abstractmethod
	def __next__(self):
		"""
		Produce the next set of data.
		"""

class Receiver(Fitting):
	_e_label = 'RECEIVER'

	@abstractmethod
	def transmit(self):
		"""
		Finish the reception of the accepted data.
		"""

	@abstractmethod
	def accept(self, data):
		"""
		Take the data object to be processed.
		"""

class NullProducer(Producer):
	"""
	Produces no copy data.
	"""
	_e_factors = ()
	protocol = PROTOCOL_NULL

	def realign(self):
		# Never needs to realigned.
		pass

	def __next__(self):
		raise StopIteration

class IteratorProducer(Producer):
	_e_factors = ('iterator',)
	protocol = PROTOCOL_CHUNKS

	def __init__(self, iterator):
		self.iterator = iter(iterator)
		self.__next__ = self.iterator.__next__
		super().__init__()

	def realign(self):
		# Never needs to realign; data is emitted on message boundaries.
		pass

	def __next__(self, next = next):
		n = next(self.iterator)
		self.total_messages += len(n)
		self.total_bytes += sum(map(len, n))
		return n

class ProtocolProducer(Producer):
	"""
	Producer using a PQv3 data stream.

	Normally, this class needs to be subclassed as it assumes that the given
	recv_into function will write COPY messages.
	"""
	protocol = PROTOCOL_PQv3

	@abstractmethod
	def recover(self, view):
		"""
		Given a view containing data read from the wire, recover the
		controller's state.

		This needs to be implemented by subclasses in order for the
		ProtocolReceiver to pass control back to the original state machine.
		"""

	##
	# When a COPY is interrupted, this can be used to accommodate
	# the original state machine to identify the message boundaries.
	def realign(self):
		s = self._state

		if s is None:
			# It's already aligned.
			self.nextchunk = iter(()).__next__
			return

		if s.final_view:
			# It was at the end or non-COPY.
			for_producer = bytes(s.final_view)
			for_receivers = b''
		elif s.remaining_bytes == -1:
			# In the middle of a message header.
			for_producer = CopyData.type + s.size_fragment
			# receivers:
			header = (self._state.size_fragment.ljust(3, b'\x00') + b'\x04')
			# Don't include the already sent parts.
			buf = header[len(self._state.size_fragment):]
			bodylen = ulong_unpack(header) - 4
			# This will often cause an invalid copy data error,
			# but it doesn't matter much because we will issue a copy fail.
			buf += b'\x00' * bodylen
			for_receivers = buf
		elif s.remaining_bytes > 0:
			# In the middle of a message.
			for_producer = CopyData.type + ulong_pack(s.remaining_bytes + 4)
			for_receivers = b'\x00' * self._state.remaining_bytes
		else:
			for_producer = for_receivers = b''

		self.recover(for_producer)
		if for_receivers:
			self.nextchunk = iter((for_receivers,)).__next__
		else:
			self.nextchunk = iter(()).__next__

	def process_copy_data(self, view):
		self.total_messages += self._state.update(view)
		if self._state.final_view is not None:
			# It's not COPY data.
			fv = self._state.final_view
			# Only publish up to the final_view.
			if fv:
				view = view[:-len(fv)]
			# The next next() will handle the async, error, or completion.
			self.recover(fv)
			self._state = None
		self.total_bytes += len(view)
		return view

	# Given a view, begin tracking the state of the wire.
	def track_state(self, view):
		self._state = WireState()
		self.nextchunk = self.recv_view
		return self.process_copy_data(view)

	# The usual method for receiving more data.
	def recv_view(self):
		view = self.buffer_view[:self.recv_into(self.buffer, self.buffer_size)]
		if not view:
			# Zero read; let the subclass handle the situation.
			self.recover(memoryview(b''))
			return self.nextchunk()
		view = self.process_copy_data(view)
		return view

	def nextchunk(self):
		raise RuntimeError("producer not properly initialized")

	def __next__(self):
		return self.nextchunk()

	def __init__(self,
		recv_into,
		buffer_size = default_buffer_size
	):
		super().__init__()
		self.recv_into = recv_into
		self.buffer_size = buffer_size
		self.buffer = bytearray(buffer_size)
		self.buffer_view = memoryview(self.buffer)
		self._state = None

class StatementProducer(ProtocolProducer):
	_e_factors = ('statement', 'parameters',)

	def _e_metas(self):
		for x in super()._e_metas():
			yield x

	@property
	def state(self):
		if self._chunks is None:
			return 'created'
		return 'producing'

	def count(self):
		return self._chunks.count()

	def command(self):
		return self._chunks.command()

	def __init__(self, statement, *args, **kw):
		super().__init__(statement.database.pq.socket.recv_into, **kw)
		self.statement = statement
		self.parameters = args
		self._chunks = None

	##
	# Take any data held by the statement's chunks and connection.
	def confiscate(self, next = next):
		current = []
		try:
			while not current:
				current.extend(next(self._chunks))
		except StopIteration:
			if not current:
				# End of COPY.
				raise
		pq = self._chunks.database.pq
		buffer = cat_messages(current) + pq.message_buffer.getvalue() + (pq.read_data or b'')
		view = memoryview(buffer)
		pq.read_data = None
		pq.message_buffer.truncate()
		# Reconstruct the buffer from the already parsed lines.
		r = self.track_state(view)
		# XXX: Better way? Probably shouldn't do the full track_state if complete..
		if self._chunks._xact.state is xactComplete:
			# It's over, don't hand off to recv_view.
			self.nextchunk = self.confiscate
			assert self._state.final_view is None
		return r

	def recover(self, view):
		# Method used when non-COPY data is found.
		self._chunks.database.pq.message_buffer.write(bytes(view))
		self.nextchunk = self.confiscate

	def __enter__(self):
		super().__enter__()
		if self._chunks is not None:
			raise RuntimeError("receiver already used")
		self._chunks = self.statement.chunks(*self.parameters)
		# Start by confiscating the connection state.
		self.nextchunk = self.confiscate

	def __exit__(self, typ, val, tb):
		if typ is None or issubclass(typ, Exception):
			db = self.statement.database
			if not db.closed and self._chunks._xact is not None:
				# The COPY transaction is still happening,
				# force an interrupt if the connection still exists.
				db.interrupt()
				if db.pq.xact:
					# Raise, CopyManager should trap.
					db._pq_complete()
		super().__exit__(typ, val, tb)

class NullReceiver(Receiver):
	_e_factors = ()
	protocol = PROTOCOL_NULL
	state = 'null'

	def transmit(self):
		# Nothing to do.
		pass

	def accept(self, data):
		pass

class ProtocolReceiver(Receiver):
	protocol = PROTOCOL_PQv3
	__slots__ = ('send', 'view')

	def __init__(self, send):
		super().__init__()
		self.send = send
		self.view = memoryview(b'')

	def accept(self, data):
		self.view = data

	def transmit(self):
		while self.view:
			self.view = self.view[self.send(self.view):]

	def __enter__(self):
		return self

	def __exit__(self, typ, val, tb):
		pass

class StatementReceiver(ProtocolReceiver):
	_e_factors = ('statement', 'parameters',)
	__slots__ = ProtocolReceiver.__slots__ + _e_factors + ('xact',)

	def _e_metas(self):
		yield None, '[' + self.state + ']'

	def __init__(self, statement, *parameters):
		self.statement = statement
		self.parameters = parameters
		self.xact = None
		super().__init__(statement.database.pq.socket.send,)

	# XXX: A bit of a hack...
	# This is actually a good indication that statements need a .copy()
	# execution method for producing a "CopyCursor" that reads or writes.
	class WireReady(BaseException):
		pass
	def raise_wire_ready(self):
		raise self.WireReady()
		yield None

	def __enter__(self, iter = iter):
		super().__enter__()
		# Get the connection in the COPY state.
		try:
			self.statement.load_chunks(
				iter(self.raise_wire_ready()), *self.parameters
			)
		except self.WireReady:
			# It's a BaseException; nothing should trap it.
			# Note the transaction object; we'll use it on exit.
			self.xact = self.statement.database.pq.xact

	def __exit__(self, typ, val, tb):
		if self.xact is None:
			# Nothing to do.
			return super().__exit__(typ, val, tb)

		if self.view:
			# The realigned producer emitted the necessary
			# data for message boundary alignment.
			#
			# In this case, we unconditionally fail.
			pq = self.statement.database.pq
			# There shouldn't be any message_data, atm.
			pq.message_data = bytes(self.view)
			self.statement.database._pq_complete()
			# It is possible for a non-alignment view to exist in cases of
			# faults. However, exit should *not* be called in those cases.
			##
		elif typ is None:
			# Success?
			self.xact.messages = self.xact.CopyDoneSequence
			# If not, this will blow up.
			self.statement.database._pq_complete()
			# Find the complete message for command and count.
			for x in self.xact.messages_received():
				if getattr(x, 'type', None) == Complete.type:
					self._complete_message = x
		elif issubclass(typ, Exception):
			# Likely raises. CopyManager should trap.
			self.statement.database._pq_complete()

		return super().__exit__(typ, val, tb)

	def count(self):
		return self._complete_message.extract_count()

	def command(self):
		return self._complete_message.extract_command().decode('ascii')

class CallReceiver(Receiver):
	"""
	Call the given object with a list of COPY lines.
	"""
	_e_factors = ('callable',)
	protocol = PROTOCOL_CHUNKS

	def __init__(self, callable):
		self.callable = callable
		self.lines = None
		super().__init__()

	def transmit(self):
		if self.lines is not None:
			self.callable(self.lines)
		self.lines = None

	def accept(self, lines):
		self.lines = lines

class CopyManager(Element, Iterator):
	"""
	A class for managing COPY operations.

	Connects the producer to the receivers.
	"""
	_e_label = 'COPY'
	_e_factors = ('producer', 'receivers',)

	def _e_metas(self):
		yield None, '[' + self.state + ']'

	@property
	def state(self):
		if self.transformer is None:
			return 'initialized'
		return str(self.producer.total_messages) + ' messages transferred'

	def __init__(self, producer, *receivers):
		self.producer = producer
		self.transformer = None
		self.receivers = ElementSet(receivers)
		self._seen_stop_iteration = False
		rp = set()
		add = rp.add
		for x in self.receivers:
			add(x.protocol)
		self.protocols = rp

	def __enter__(self):
		if self.transformer:
			raise RuntimeError("copy already started")
		self._stats = (0, 0)
		self.transformer = CopyTransformer(self.producer.protocol, self.protocols)
		self.producer.__enter__()
		try:
			for x in self.receivers:
				x.__enter__()
		except Exception:
			self.__exit__(*sys.exc_info())
		return self

	def __exit__(self, typ, val, tb):
		##
		# Exiting the CopyManager is a fairly complex operation.
		#
		# In cases of failure, re-alignment may need to happen
		# for when the receivers are not on a message boundary.
		##
		if typ is not None and not issubclass(typ, Exception):
			# Don't bother, it's an interrupt or sufficient resources.
			return

		profail = None
		try:
			# Does nothing if the COPY was successful.
			self.producer.realign()
			try:
				##
				# If the producer is not aligned to a message boundary,
				# it can emit completion data that will put the receivers
				# back on track.
				# This last service call will move that data onto the receivers.
				self._service_producer()
				##
				# The receivers need to handle any new data in their __exit__.
			except StopIteration:
				# No re-alignment needed.
				pass

			self.producer.__exit__(typ, val, tb)
		except Exception as x:
			# reference profail later.
			profail = x

		# No receivers? It wasn't a success.
		if not self.receivers:
			raise CopyFail(self, "no receivers", producer_fault = profail)

		exit_faults = {}
		for x in self.receivers:
			try:
				x.__exit__(typ, val, tb)
			except Exception as e:
				exit_faults[x] = e

		if typ or exit_faults or profail or not self._seen_stop_iteration:
			raise CopyFail(self,
				"could not complete the COPY operation",
				receiver_faults = exit_faults or None,
				producer_fault = profail
			)

	def reconcile(self, r):
		"""
		Reconcile a receiver that faulted.

		This method should be used to add back a receiver that failed to
		complete its write operation, but is capable of completing the
		operation at this time.
		"""
		if r.protocol not in self.protocols:
			raise RuntimeError("cannot add new receivers to copy operations")
		r.transmit()
		# Okay, add it back.
		self.receivers.add(r)

	def _service_producer(self):
		# Setup current data.
		if not self.receivers:
			# No receivers to take the data.
			raise StopIteration

		try:
			nextdata = next(self.producer)
		except StopIteration:
			# Should be over.
			self._seen_stop_iteration = True
			raise
		except Exception:
			raise ProducerFault(self)

		self.transformer(nextdata)

		# Distribute data to receivers.
		for x in self.receivers:
			x.accept(self.transformer.get(x.protocol))

	def _service_receivers(self):
		faults = {}
		for x in self.receivers:
			# Process all the receivers.
			try:
				x.transmit()
			except Exception as e:
				faults[x] = e
		if faults:
			# The CopyManager is eager to continue the operation.
			for x in faults:
				self.receivers.discard(x)
			raise ReceiverFault(self, faults)

	# Run the COPY to completion.
	def run(self):
		with self:
			try:
				while True:
					self._service_producer()
					self._service_receivers()
			except StopIteration:
				# It's done.
				pass

	def __iter__(self):
		return self

	def __next__(self):
		messages = self.producer.total_messages
		bytes = self.producer.total_bytes

		self._service_producer()
		# Record the progress in case a receiver faults.
		self._stats = (
			self._stats[0] + (self.producer.total_messages - messages),
			self._stats[1] + (self.producer.total_bytes - bytes),
		)
		self._service_receivers()
		# Return the progress.
		current_stats = self._stats
		self._stats = (0, 0)
		return current_stats

def transfer(producer, *receivers):
	"""
	Perform a COPY operation using the given statements::

		>>> import copyman
		>>> copyman.transfer(src.prepare("COPY table TO STDOUT"), dst.prepare("COPY table FROM STDIN"))
	"""
	cm = CopyManager(
		StatementProducer(producer),
		*[x if isinstance(x, Receiver) else StatementReceiver(x) for x in receivers]
	)
	cm.run()
	return (cm.producer.total_messages, cm.producer.total_bytes)



##
# .python.element
##
import os
from abc import ABCMeta, abstractproperty, abstractmethod
from .string import indent
from .decorlib import propertydoc

class RecursiveFactor(Exception):
	"""
	Raised when a factor is ultimately composed of itself.
	"""
	pass

class Element(object, metaclass = ABCMeta):
	"""
	The purpose of an element is to provide a general mechanism for specifying
	the factors that composed an object. Factors are designated using an
	ordered set of strings referencing those significant attributes on the object.

	Factors are important for PG-API as it provides the foundation for
	collecting the information about the state of the interface that ultimately
	led up to an error.

		Traceback:
		 ...
		postgresql.exceptions.*: <message>
		  CODE: XX000
		CURSOR: <cursor_id>
		  parameters: (p1, p2, ...)
		STATEMENT: <statement_id> <parameter info>
		  ...
		  string:
		    <query body>
		  SYMBOL: get_types
		  LIBRARY: catalog
		  ...
		CONNECTION:
		  <backend_id> <socket information>
		CONNECTOR: [Host]
		  IRI: pq://user@localhost:5432/database
		DRIVER: postgresql.driver.pq3
	"""

	@propertydoc
	@abstractproperty
	def _e_label(self) -> str:
		"""
		Single-word string describing the kind of element.

		For instance, `postgresql.api.Statement`'s _e_label is 'STATEMENT'.

		Usually, this is set directly on the class itself, and is a shorter
		version of the class's name.
		"""

	@propertydoc
	@abstractproperty
	def _e_factors(self) -> ():
		"""
		The attribute names of the objects that contributed to the creation of
		this object.

		The ordering is significant. The first factor is the prime factor.
		"""

	@abstractmethod
	def _e_metas(self) -> [(str, object)]:
		"""
		Return an iterable to key-value pairs that provide useful descriptive
		information about an attribute.

		Factors on metas are not checked. They are expected to be primitives.

		If there are no metas, the str() of the object will be used to represent
		it.
		"""

class ElementSet(Element, set):
	"""
	An ElementSet is a set of Elements that can be used as an individual factor.

	In situations where a single factor is composed of multiple elements where
	each has no significance over the other, this Element can be used represent
	that fact.

	Importantly, it provides the set metadata so that the appropriate information
	will be produced in element tracebacks.
	"""
	_e_label = 'SET'
	_e_factors = ()
	__slots__ = ()

	def _e_metas(self):
		yield (None, len(self))
		for x in self:
			yield (None, '--')
			yield (None, format_element(x))

def prime_factor(obj):
	"""
	Get the primary factor on the `obj`, returns None if none.
	"""
	f = getattr(obj, '_e_factors', None)
	if f:
		return f[0], getattr(obj, f[0], None)

def prime_factors(obj):
	"""
	Yield out the sequence of primary factors of the given object.
	"""
	visited = set((obj,))
	ef = getattr(obj, '_e_factors', None)
	if not ef:
		return
	fn = ef[0]
	e = getattr(obj, fn, None)
	if e in visited:
		raise RecursiveFactor(obj, e)
	visited.add(e)
	yield fn, e

	while e is not None:
		ef = getattr(obj, '_e_factors', None)
		fn = ef[0]
		e = getattr(e, fn, None)
		if e in visited:
			raise RecursiveFactor(obj, e)
		visited.add(e)
		yield fn, e

def format_element(obj, coverage = ()):
	"""
	Format the given element with its factors and metadata into a readable string.
	"""
	# if it's not an Element, all there is to return is str(obj)
	if obj in coverage:
		raise RecursiveFactor(coverage)
	coverage = coverage + (obj,)

	if not isinstance(obj, Element):
		if obj is None:
			return 'None'
		return str(obj)

	# The description of `obj` is built first.

	# formal element, get metas first.
	nolead = False
	metas = []
	for key, val in obj._e_metas():
		m = ""
		if val is None:
			sval = 'None'
		else:
			sval = str(val)

		pre = ' '
		if key is not None:
			m += key + ':'
			if (len(sval) > 70 or os.linesep in sval):
				pre = os.linesep
				sval = indent(sval)
		else:
			# if the key is None, it is intended to be inlined.
			nolead = True
			pre = ''
		m += pre + sval.rstrip()
		metas.append(m)

	factors = []
	for att in obj._e_factors[1:]:
		m = ""
		f = getattr(obj, att)
		# if the object has a label, use the label
		m += att + ':'
		sval = format_element(f, coverage = coverage)
		if len(sval) > 70 or os.linesep in sval:
			m += os.linesep + indent(sval)
		else:
			m += ' ' + sval
		factors.append(m)

	mtxt = os.linesep.join(metas)
	ftxt = os.linesep.join(factors)
	if mtxt:
		mtxt = indent(mtxt)
	if ftxt:
		ftxt = indent(ftxt)
	s = mtxt + ftxt
	if nolead is True:
		# metas started with a `None` key.
		s = ' ' + s.lstrip()
	else:
		s = os.linesep + s
	s = obj._e_label + ':' + s.rstrip()

	# and resolve the next prime
	pf = prime_factor(obj)
	if pf is not None:
		factor_name, prime = pf
		factor = format_element(prime, coverage = coverage)
		if getattr(prime, '_e_label', None) is not None:
			# if the factor has a label, then it will be
			# included in the format_element output, and
			# thus factor_name is not needed.
			factor_name = ''
		else:
			factor_name += ':'
			if len(factor) > 70 or os.linesep in factor:
				factor = os.linesep + indent(factor)
			else:
				factor_name += ' '
		s += os.linesep + factor_name + factor
	return s



import struct
from math import floor
from ...python.functools import Composition as compose
from ...python.itertools import interlace
from ...python.structlib import \
	short_pack, short_unpack, \
	ulong_pack, ulong_unpack, \
	long_pack, long_unpack, \
	double_pack, double_unpack, \
	longlong_pack, longlong_unpack, \
	float_pack, float_unpack, \
	LH_pack, LH_unpack, \
	dl_pack, dl_unpack, \
	dll_pack, dll_unpack, \
	ql_pack, ql_unpack, \
	qll_pack, qll_unpack, \
	llL_pack, llL_unpack, \
	dd_pack, dd_unpack, \
	ddd_pack, ddd_unpack, \
	dddd_pack, dddd_unpack, \
	hhhh_pack, hhhh_unpack

oid_pack = cid_pack = xid_pack = ulong_pack
oid_unpack = cid_unpack = xid_unpack = ulong_unpack
tid_pack, tid_unpack = LH_pack, LH_unpack

# geometry types
point_pack, point_unpack = dd_pack, dd_unpack
circle_pack, circle_unpack = ddd_pack, ddd_unpack
lseg_pack = box_pack = dddd_pack
lseg_unpack = box_unpack = dddd_unpack

null_sequence = b'\xff\xff\xff\xff'
string_format = b'\x00\x00'
binary_format = b'\x00\x01'

def numeric_pack(data, hhhh_pack = hhhh_pack, pack = struct.pack, len = len):
	return hhhh_pack(data[0]) + pack("!%dh"%(len(data[1]),), *data[1])

def numeric_unpack(data, hhhh_unpack = hhhh_unpack, unpack = struct.unpack, len = len):
	return (hhhh_unpack(data[:8]), unpack("!8x%dh"%((len(data)-8) // 2,), data))

def path_pack(data, pack = struct.pack, len = len):
	"""
	Given a sequence of point data, pack it into a path's serialized form.

		[px1, py1, px2, py2, ...]

	Must be an even number of numbers.
	"""
	return pack("!l%dd" %(len(data),), len(data), *data)

def path_unpack(data, long_unpack = long_unpack, unpack = struct.unpack):
	"""
	Unpack a path's serialized form into a sequence of point data:

		[px1, py1, px2, py2, ...]

	Should be an even number of numbers.
	"""
	return unpack("!4x%dd" %(long_unpack(data[:4]),), data)
polygon_pack, polygon_unpack = path_pack, path_unpack

##
# Binary representations of infinity for datetimes.
time_infinity = b'\x7f\xf0\x00\x00\x00\x00\x00\x00'
time_negative_infinity = b'\xff\xf0\x00\x00\x00\x00\x00\x00'
time64_infinity = b'\x7f\xff\xff\xff\xff\xff\xff\xff'
time64_negative_infinity = b'\x80\x00\x00\x00\x00\x00\x00\x00'
date_infinity = b'\x7f\xff\xff\xff'
date_negative_infinity = b'\x80\x00\x00\x00'

# time types
date_pack, date_unpack = long_pack, long_unpack

def mktimetuple(ts, floor = floor):
	'make a pair of (seconds, microseconds) out of the given double'
	seconds = floor(ts)
	return (int(seconds), int(1000000 * (ts - seconds)))

def mktimetuple64(ts, divmod = divmod):
	'make a pair of (seconds, microseconds) out of the given long'
	return divmod(ts, 1000000)

def mktime(seconds_ms, float = float):
	'make a double out of the pair of (seconds, microseconds)'
	return float(seconds_ms[0]) + (seconds_ms[1] / 1000000.0)

def mktime64(seconds_ms):
	'make an integer out of the pair of (seconds, microseconds)'
	return seconds_ms[0] * 1000000 + seconds_ms[1]

# takes a pair, (seconds, microseconds)
time_pack = compose((mktime, double_pack))
time_unpack = compose((double_unpack, mktimetuple))

def interval_pack(m_d_timetup, mktime = mktime, dll_pack = dll_pack):
	"""
	Given a triple, (month, day, (seconds, microseconds)), serialize it for
	transport.
	"""
	(month, day, timetup) = m_d_timetup
	return dll_pack((mktime(timetup), day, month))

def interval_unpack(data, dll_unpack = dll_unpack, mktimetuple = mktimetuple):
	"""
	Given a serialized interval, '{month}{day}{time}', yield the triple:

		(month, day, (seconds, microseconds))
	"""
	tim, day, month = dll_unpack(data)
	return (month, day, mktimetuple(tim))

def interval_noday_pack(month_day_timetup, dl_pack = dl_pack, mktime = mktime):
	"""
	Given a triple, (month, day, (seconds, microseconds)), return the serialized
	form that does not have an individual day component.

	There is no day component, so if day is non-zero, it will be converted to
	seconds and subsequently added to the seconds.
	"""
	(month, day, timetup) = month_day_timetup
	if day:
		timetup = (timetup[0] + (day * 24 * 60 * 60), timetup[1])
	return dl_pack((mktime(timetup), month))

def interval_noday_unpack(data, dl_unpack = dl_unpack, mktimetuple = mktimetuple):
	"""
	Given a serialized interval without a day component, return the triple:

		(month, day(always zero), (seconds, microseconds))
	"""
	tim, month = dl_unpack(data)
	return (month, 0, mktimetuple(tim))

def time64_pack(data, mktime64 = mktime64, longlong_pack = longlong_pack):
	return longlong_pack(mktime64(data))
def time64_unpack(data, longlong_unpack = longlong_unpack, mktimetuple64 = mktimetuple64):
	return mktimetuple64(longlong_unpack(data))

def interval64_pack(m_d_timetup, qll_pack = qll_pack, mktime64 = mktime64):
	"""
	Given a triple, (month, day, (seconds, microseconds)), return the serialized
	data using a quad-word for the (seconds, microseconds) tuple.
	"""
	(month, day, timetup) = m_d_timetup
	return qll_pack((mktime64(timetup), day, month))

def interval64_unpack(data, qll_unpack = qll_unpack, mktimetuple = mktimetuple):
	"""
	Unpack an interval containing a quad-word into a triple:

		(month, day, (seconds, microseconds))
	"""
	tim, day, month = qll_unpack(data)
	return (month, day, mktimetuple64(tim))

def interval64_noday_pack(m_d_timetup, ql_pack = ql_pack, mktime64 = mktime64):
	"""
	Pack an interval without a day component and using a quad-word for second
	representation.

	There is no day component, so if day is non-zero, it will be converted to
	seconds and subsequently added to the seconds.
	"""
	(month, day, timetup) = m_d_timetup
	if day:
		timetup = (timetup[0] + (day * 24 * 60 * 60), timetup[1])
	return ql_pack((mktime64(timetup), month))

def interval64_noday_unpack(data, ql_unpack = ql_unpack, mktimetuple64 = mktimetuple64):
	"""
	Unpack a ``noday`` quad-word based interval. Returns a triple:

		(month, day(always zero), (seconds, microseconds))
	"""
	tim, month = ql_unpack(data)
	return (month, 0, mktimetuple64(tim))

def timetz_pack(timetup_tz, dl_pack = dl_pack, mktime = mktime):
	"""
	Pack a time; offset from beginning of the day and timezone offset.

	Given a pair, ((seconds, microseconds), timezone_offset), pack it into its
	serialized form: "!dl".
	"""
	(timetup, tz_offset) = timetup_tz
	return dl_pack((mktime(timetup), tz_offset))

def timetz_unpack(data, dl_unpack = dl_unpack, mktimetuple = mktimetuple):
	"""
	Given serialized time data, unpack it into a pair:

	    ((seconds, microseconds), timezone_offset).
	"""
	ts, tz = dl_unpack(data)
	return (mktimetuple(ts), tz)

def timetz64_pack(timetup_tz, ql_pack = ql_pack, mktime64 = mktime64):
	"""
	Pack a time; offset from beginning of the day and timezone offset.

	Given a pair, ((seconds, microseconds), timezone_offset), pack it into its
	serialized form using a long long: "!ql".
	"""
	(timetup, tz_offset) = timetup_tz
	return ql_pack((mktime64(timetup), tz_offset))

def timetz64_unpack(data, ql_unpack = ql_unpack, mktimetuple64 = mktimetuple64):
	"""
	Given "long long" serialized time data, "ql", unpack it into a pair:

	    ((seconds, microseconds), timezone_offset)
	"""
	ts, tz = ql_unpack(data)
	return (mktimetuple64(ts), tz)

# oidvectors are 128 bytes, so pack the number of Oids in self
# and justify that to 128 by padding with \x00.
def oidvector_pack(seq, pack = struct.pack):
	"""
	Given a sequence of Oids, pack them into the serialized form.

	An oidvector is a type used by the PostgreSQL catalog.
	"""
	return pack("!%dL"%(len(seq),), *seq).ljust(128, '\x00')

def oidvector_unpack(data, unpack = struct.unpack):
	"""
	Given a serialized oidvector(32 longs), unpack it into a list of unsigned integers.

	An int2vector is a type used by the PostgreSQL catalog.
	"""
	return unpack("!32L", data)

def int2vector_pack(seq, pack = struct.pack):
	"""
	Given a sequence of integers, pack them into the serialized form.

	An int2vector is a type used by the PostgreSQL catalog.
	"""
	return pack("!%dh"%(len(seq),), *seq).ljust(64, '\x00')

def int2vector_unpack(data, unpack = struct.unpack):
	"""
	Given a serialized int2vector, unpack it into a list of integers.

	An int2vector is a type used by the PostgreSQL catalog.
	"""
	return unpack("!32h", data)

def varbit_pack(bits_data, long_pack = long_pack):
	r"""
	Given a pair, serialize the varbit.

	# (number of bits, data)
	>>> varbit_pack((1, '\x00'))
	b'\x00\x00\x00\x01\x00'
	"""
	return long_pack(bits_data[0]) + bits_data[1]

def varbit_unpack(data, long_unpack = long_unpack):
	"""
	Given ``varbit`` data, unpack it into a pair:

		(bits, data)

	Where bits are the total number of bits in data (bytes).
	"""
	return long_unpack(data[0:4]), data[4:]

def net_pack(triple,
	# Map PGSQL src/include/utils/inet.h to IP version number.
	fmap = {
		4: 2,
		6: 3,
	},
	len = len,
):
	"""
	net_pack((family, mask, data))

	Pack Postgres' inet/cidr data structure.
	"""
	family, mask, data = triple
	return bytes((fmap[family], mask or 0, 0 if mask is None else 1, len(data))) + data

def net_unpack(data,
	# Map IP version number to PGSQL src/include/utils/inet.h.
	fmap = {
		2: 4,
		3: 6,
	}
):
	"""
	net_unpack(data)

	Unpack Postgres' inet/cidr data structure.
	"""
	family, mask, is_cidr, size = data[:4]
	return (fmap[family], mask, data[4:])

def macaddr_pack(data, bytes = bytes):
	"""
	Pack a MAC address

	Format found in PGSQL src/backend/utils/adt/mac.c, and PGSQL Manual types
	"""
	# Accept all possible PGSQL Macaddr formats as in manual
	# Oh for sscanf() as we could just copy PGSQL C in src/util/adt/mac.c
	colon_parts = data.split(':')
	dash_parts = data.split('-')
	dot_parts = data.split('.')
	if len(colon_parts) == 6:
		mac_parts = colon_parts
	elif len(dash_parts) == 6:
		mac_parts = dash_parts
	elif len(colon_parts) == 2:
		mac_parts = [colon_parts[0][:2], colon_parts[0][2:4], colon_parts[0][4:],
			colon_parts[1][:2], colon_parts[1][2:4], colon_parts[1][4:]]
	elif len(dash_parts) == 2:
		mac_parts = [dash_parts[0][:2], dash_parts[0][2:4], dash_parts[0][4:],
			dash_parts[1][:2], dash_parts[1][2:4], dash_parts[1][4:]]
	elif len(dot_parts) == 3:
		mac_parts = [dot_parts[0][:2], dot_parts[0][2:], dot_parts[1][:2],
			dot_parts[1][2:], dot_parts[2][:2], dot_parts[2][2:]]
	elif len(colon_parts) == 1:
		mac_parts = [data[:2], data[2:4], data[4:6], data[6:8], data[8:10], data[10:]]
	else:
		raise ValueError('data string cannot be parsed to bytes')
	if len(mac_parts) != 6 and len(mac_parts[-1]) != 2:
		raise ValueError('data string cannot be parsed to bytes')
	return bytes([int(p, 16) for p in mac_parts])

def macaddr_unpack(data):
	"""
	Unpack a MAC address

	Format found in PGSQL src/backend/utils/adt/mac.c
	"""
	# This is easy, just go for standard macaddr format,
	# just like PGSQL in src/util/adt/mac.c macaddr_out()
	if len(data) != 6:
		raise ValueError('macaddr has incorrect length')
	return ("%02x:%02x:%02x:%02x:%02x:%02x" % tuple(data))

def record_unpack(data,
	long_unpack = long_unpack,
	oid_unpack = oid_unpack,
	null_sequence = null_sequence,
	len = len):
	"""
	Given serialized record data, return a tuple of tuples of type Oids and
	attributes.
	"""
	columns = long_unpack(data)
	offset = 4

	for x in range(columns):
		typid = oid_unpack(data[offset:offset+4])
		offset += 4

		if data[offset:offset+4] == null_sequence:
			att = None
			offset += 4
		else:
			size = long_unpack(data[offset:offset+4])
			offset += 4
			att = data[offset:offset + size]
			if size < -1 or len(att) != size:
				raise ValueError("insufficient data left in message")
			offset += size
		yield (typid, att)

	if len(data) - offset != 0:
		raise ValueError("extra data, %d octets, at end of record" %(len(data),))

def record_pack(seq,
	long_pack = long_pack,
	oid_pack = oid_pack,
	null_sequence = null_sequence):
	"""
	pack a record given an iterable of (type_oid, data) pairs.
	"""
	return long_pack(len(seq)) + b''.join([
		# typid + (null_seq or data)
		oid_pack(x) + (y is None and null_sequence or (long_pack(len(y)) + y))
		for x, y in seq
	])

def elements_pack(elements,
	null_sequence = null_sequence,
	long_pack = long_pack, len = len
):
	"""
	Pack the elements for containment within a serialized array.

	This is used by array_pack.
	"""
	for x in elements:
		if x is None:
			yield null_sequence
		else:
			yield long_pack(len(x))
			yield x

def array_pack(array_data,
	llL_pack = llL_pack,
	len = len,
	long_pack = long_pack,
	interlace = interlace
):
	"""
	Pack a raw array. A raw array consists of flags, type oid, sequence of lower
	and upper bounds, and an iterable of already serialized element data:

		(0, element type oid, (lower bounds, upper bounds, ...), iterable of element_data)

	The lower bounds and upper bounds specifies boundaries of the dimension. So the length
	of the boundaries sequence is two times the number of dimensions that the array has.

	array_pack((flags, type_id, dims, lowers, element_data))

	The format of ``lower_upper_bounds`` is a sequence of lower bounds and upper
	bounds. First lower then upper inlined within the sequence:

		[lower, upper, lower, upper]

	The above array `dlb` has two dimensions. The lower and upper bounds of the
	first dimension is defined by the first two elements in the sequence. The
	second dimension is then defined by the last two elements in the sequence.
	"""
	(flags, typid, dims, lbs, elements) = array_data
	return llL_pack((len(dims), flags, typid)) + \
		b''.join(map(long_pack, interlace(dims, lbs))) + \
		b''.join(elements_pack(elements))

def elements_unpack(data, offset,
	long_unpack = long_unpack,
	null_sequence = null_sequence):
	"""
	Unpack the serialized elements of an array into a list.

	This is used by array_unpack.
	"""
	data_len = len(data)
	while offset < data_len:
		lend = data[offset:offset+4]
		offset += 4
		if lend == null_sequence:
			yield None
		else:
			sizeof_el = long_unpack(lend)
			yield data[offset:offset+sizeof_el]
			offset += sizeof_el

def array_unpack(data,
	llL_unpack = llL_unpack,
	unpack = struct.unpack_from,
	long_unpack = long_unpack
):
	"""
	Given a serialized array, unpack it into a tuple:

		(flags, typid, (dims, lower bounds, ...), [elements])
	"""
	ndim, flags, typid = llL_unpack(data)
	if ndim < 0:
		raise ValueError("invalid number of dimensions: %d" %(ndim,))
	# "ndim" number of pairs of longs
	end = (4 * 2 * ndim) + 12
	# Dimensions and lower bounds; split the two early.
	#dlb = unpack("!%dl"%(2 * ndim,), data, 12)
	dims = [long_unpack(data[x:x+4]) for x in range(12, end, 8)]
	lbs = [long_unpack(data[x:x+4]) for x in range(16, end, 8)]
	return (flags, typid, dims, lbs, elements_unpack(data, end))



#!/usr/bin/env python
##
# setup.py - .release.distutils
##
import sys
import os

if sys.version_info[:2] < (3,3):
	sys.stderr.write(
		"ERROR: py-postgresql is for Python 3.3 and greater." + os.linesep
	)
	sys.stderr.write(
		"HINT: setup.py was ran using Python " + \
		'.'.join([str(x) for x in sys.version_info[:3]]) +
		': ' + sys.executable + os.linesep
	)
	sys.exit(1)

# distutils data is kept in `postgresql.release.distutils`
sys.path.insert(0, '')

sys.dont_write_bytecode = True
import py_opengauss.release.distutils as dist
defaults = dist.standard_setup_keywords()
sys.dont_write_bytecode = False

if __name__ == '__main__':
	try:
		from setuptools import setup
	except ImportError as e:
		from distutils.core import setup
	setup(**defaults)



from .. import POINTOID, BOXOID, LSEGOID, CIRCLEOID
from ..geometry import Point, Box, Lseg, Circle
from ...python.functools import Composition as compose
from . import lib

oid_to_type = {
	POINTOID: Point,
	BOXOID: Box,
	LSEGOID: Lseg,
	CIRCLEOID: Circle,
}

# Make a pair of pairs out of a sequence of four objects
def two_pair(x):
	return ((x[0], x[1]), (x[2], x[3]))

point_pack = lib.point_pack
point_unpack = compose((lib.point_unpack, Point))

def box_pack(x):
	return lib.box_pack((x[0][0], x[0][1], x[1][0], x[1][1]))
box_unpack = compose((lib.box_unpack, two_pair, Box,))

def lseg_pack(x, pack = lib.lseg_pack):
	return pack((x[0][0], x[0][1], x[1][0], x[1][1]))
lseg_unpack = compose((lib.lseg_unpack, two_pair, Lseg))

def circle_pack(x):
	return lib.circle_pack((x[0][0], x[0][1], x[1]))
def circle_unpack(x, unpack = lib.circle_unpack, Circle = Circle):
	x = unpack(x)
	return Circle(((x[0], x[1]), x[2]))

# Map type oids to a (pack, unpack) pair.
oid_to_io = {
	POINTOID : (point_pack, point_unpack, Point),
	BOXOID : (box_pack, box_unpack, Box),
	LSEGOID : (lseg_pack, lseg_unpack, Lseg),
	CIRCLEOID : (circle_pack, circle_unpack, Circle),
	#PATHOID : (path_pack, path_unpack),
	#POLYGONOID : (polygon_pack, polygon_unpack),
	#LINEOID : (line_pack, line_unpack),
}



##
# .test.support
##
"""
Executable module used by test_* modules to mimic a command.
"""
import sys

def pg_config(*args):
	data = """FOO=BaR
FEH=YEAH
version=NAY
"""
	sys.stdout.write(data)

if __name__ == '__main__':
	if sys.argv[1:]:
		cmd = sys.argv[1]
		if cmd in globals():
			cmd = globals()[cmd]
			cmd(sys.argv[2:])
			sys.exit(0)
	sys.stderr.write("no valid entry point referenced")
	sys.exit(1)



##
# .bin.pg_python - Python console with a connection.
##
"""
Python command with a PG-API connection(``db``).
"""
import os
import sys
import re
import code
import optparse
import contextlib
from .. import clientparameters
from ..python import command as pycmd
from .. import project

from ..driver import default as pg_driver
from .. import exceptions as pg_exc
from .. import sys as pg_sys
from .. import lib as pg_lib

pq_trace = optparse.make_option(
	'--pq-trace',
	dest = 'pq_trace',
	help = 'trace PQ protocol transmissions',
	default = None,
)
default_options = [
	pq_trace,
	clientparameters.option_lib,
	clientparameters.option_libpath,
] + pycmd.default_optparse_options

def command(argv = sys.argv):
	p = clientparameters.DefaultParser(
		"%prog [connection options] [script] ...",
		version = project.version,
		option_list = default_options
	)
	p.disable_interspersed_args()
	co, ca = p.parse_args(argv[1:])
	rv = 1

	# Resolve the category.
	pg_sys.libpath.insert(0, os.path.curdir)
	pg_sys.libpath.extend(co.libpath or [])
	if co.lib:
		cat = pg_lib.Category(*map(pg_lib.load, co.lib))
	else:
		cat = None

	trace_file = None
	if co.pq_trace is not None:
		trace_file = open(co.pq_trace, 'a')

	try:
		need_prompt = False
		cond = None
		connector = None
		connection = None
		while connection is None:
			try:
				cond = clientparameters.collect(parsed_options = co, prompt_title = None)
				if need_prompt:
					# authspec error thrown last time, so force prompt.
					cond['prompt_password'] = True
				try:
					clientparameters.resolve_password(cond, prompt_title = 'pg_python')
				except EOFError:
					raise SystemExit(1)
				connector = pg_driver.fit(category = cat, **cond)
				connection = connector()
				if trace_file is not None:
					connection.tracer = trace_file.write
				connection.connect()
			except pg_exc.ClientCannotConnectError as err:
				for att in connection.failures:
					exc = att.error
					if isinstance(exc, pg_exc.AuthenticationSpecificationError):
						sys.stderr.write(os.linesep + exc.message + (os.linesep*2))
						# keep prompting the user
						need_prompt = True
						connection = None
						break
				else:
					# no invalid password failures..
					raise

		pythonexec = pycmd.Execution(ca,
			context = getattr(co, 'python_context', None),
			loader = getattr(co, 'python_main', None),
		)

		builtin_overload = {
		# New built-ins
			'connector' : connector,
			'db' : connection,
			'do' : connection.do,
			'prepare' : connection.prepare,

			'sqlexec' : connection.execute,
			'settings' : connection.settings,
			'proc' : connection.proc,
			'xact' : connection.xact,
		}
		if not isinstance(__builtins__, dict):
			builtins_d = __builtins__.__dict__
		else:
			builtins_d = __builtins__
		restore = {k : builtins_d.get(k) for k in builtin_overload}

		builtins_d.update(builtin_overload)
		try:
			with connection:
				rv = pythonexec(
					context = pycmd.postmortem(os.environ.get('PYTHON_POSTMORTEM'))
				)
				exc = getattr(sys, 'last_type', None)
				if rv and exc and not issubclass(exc, Exception):
					# Don't try to close it if wasn't an Exception.
					del connection.pq.socket
		finally:
			# restore __builtins__
			builtins_d.update(restore)
			for k, v in builtin_overload.items():
				if v is None:
					del builtins_d[x]
			if trace_file is not None:
				trace_file.close()
	except:
		pg_sys.libpath.remove(os.path.curdir)
		raise
	return rv

if __name__ == '__main__':
	sys.exit(command(sys.argv))



##
# .python.doc
##
"""
Documentation Tools.
"""
from operator import attrgetter

class Doc(object):
	"""
	Simple object that sets the __doc__ attribute to the first parameter and
	initializes __annotations__ using keyword arguments.
	"""
	def __init__(self, doc, **annotations):
		self.__doc__ = str(doc)
		self.__annotations__ = annotations

	__str__ = attrgetter('__doc__')



##
# .test.test_copyman - test .copyman
##
import unittest
from itertools import islice
from .. import copyman
from ..temporal import pg_tmp
# The asyncs, and alternative termination.
from ..protocol.element3 import Notice, Notify, Error, cat_messages
from .. import exceptions as pg_exc

# state manager can handle empty data messages, right? =)
emptysource = """
CREATE TEMP TABLE emptysource ();
-- 10
INSERT INTO emptysource DEFAULT VALUES;
INSERT INTO emptysource DEFAULT VALUES;
INSERT INTO emptysource DEFAULT VALUES;
INSERT INTO emptysource DEFAULT VALUES;
INSERT INTO emptysource DEFAULT VALUES;
INSERT INTO emptysource DEFAULT VALUES;
INSERT INTO emptysource DEFAULT VALUES;
INSERT INTO emptysource DEFAULT VALUES;
INSERT INTO emptysource DEFAULT VALUES;
INSERT INTO emptysource DEFAULT VALUES;
"""
emptydst = "CREATE TEMP TABLE empty ();"

# The usual subjects.
stdrowcount = 10000
stdsource = """
CREATE TEMP TABLE source (i int, t text);
INSERT INTO source
	SELECT i, i::text AS t
	FROM generate_series(1, {0}) AS g(i);
""".format(stdrowcount)
stditer = [
	b'\t'.join((x, x)) + b'\n'
	for x in (
		str(i).encode('ascii') for i in range(1, 10001)
	)
]
stditer_tuples = [
	(x, str(x)) for x in range(1, 10001)
]

stddst = "CREATE TEMP TABLE destination (i int, t text)"
srcsql = "COPY source TO STDOUT"
dstsql = "COPY destination FROM STDIN"
binary_srcsql = "COPY source TO STDOUT WITH BINARY"
binary_dstsql = "COPY destination FROM STDIN WITH BINARY"
dstcount = "SELECT COUNT(*) FROM destination"
grabdst = "SELECT * FROM destination ORDER BY i ASC"
grabsrc = "SELECT * FROM source ORDER BY i ASC"

##
# This subclass is used to append some arbitrary data
# after the initial data. This is used to exercise async/notice support.
class Injector(copyman.StatementProducer):
	def __init__(self, appended_messages, *args, **kw):
		super().__init__(*args, **kw)
		self._appended_messages = appended_messages

	def confiscate(self):
		pq = self.statement.database.pq
		mb = pq.message_buffer
		b = mb.getvalue()
		mb.truncate()
		mb.write(cat_messages(self._appended_messages))
		mb.write(b)
		return super().confiscate()

class test_copyman(unittest.TestCase):
	def testNull(self):
		# Test some of the basic machinery.
		sp = copyman.NullProducer()
		sr = copyman.NullReceiver()
		copyman.CopyManager(sp, sr).run()
		self.assertEqual(sp.total_messages, 0)
		self.assertEqual(sp.total_bytes, 0)

	@pg_tmp
	def testNullProducer(self):
		sqlexec(stddst)
		np = copyman.NullProducer()
		sr = copyman.StatementReceiver(prepare(dstsql))
		copyman.CopyManager(np, sr).run()
		self.assertEqual(np.total_messages, 0)
		self.assertEqual(np.total_bytes, 0)
		self.assertEqual(prepare(dstcount).first(), 0)
		self.assertEqual(prepare(grabdst)(), [])

	@pg_tmp
	def testNullReceiver(self):
		sqlexec(stdsource)
		sp = copyman.StatementProducer(prepare(srcsql), buffer_size = 128)
		sr = copyman.NullReceiver()
		with copyman.CopyManager(sp, sr) as copy:
			for x in copy:
				pass
		self.assertEqual(sp.total_messages, stdrowcount)
		self.assertEqual(sp.total_bytes > 0, True)

	def testIteratorToCall(self):
		tmp = iter(stditer)
		# segment stditer into chunks consisting of twenty rows each
		sp = copyman.IteratorProducer([
			list(islice(tmp, 20)) for x in range(len(stditer) // 20)
		])
		dest = []
		sr = copyman.CallReceiver(dest.extend)
		recomputed_bytes = 0
		recomputed_messages = 0
		with copyman.CopyManager(sp, sr) as copy:
			for msg, bytes in copy:
				recomputed_messages += msg
				recomputed_bytes += bytes
		self.assertEqual(stdrowcount, recomputed_messages)
		self.assertEqual(recomputed_bytes, sp.total_bytes)
		self.assertEqual(len(dest), stdrowcount)
		self.assertEqual(dest, stditer)

	@pg_tmp
	def testDirectStatements(self):
		sqlexec(stdsource)
		dst = new()
		dst.execute(stddst)
		sp = copyman.StatementProducer(prepare(srcsql), buffer_size = 512)
		sr = copyman.StatementReceiver(dst.prepare(dstsql))
		with copyman.CopyManager(sp, sr) as copy:
			for x in copy:
				pass
		self.assertEqual(dst.prepare(dstcount).first(), stdrowcount)
		self.assertEqual(dst.prepare(grabdst)(), prepare(grabsrc)())

	@pg_tmp
	def testIteratorProducer(self):
		sqlexec(stddst)
		sp = copyman.IteratorProducer([stditer])
		sr = copyman.StatementReceiver(prepare(dstsql))
		recomputed_bytes = 0
		recomputed_messages = 0
		with copyman.CopyManager(sp, sr) as copy:
			for msg, bytes in copy:
				recomputed_messages += msg
				recomputed_bytes += bytes
		self.assertEqual(stdrowcount, recomputed_messages)
		self.assertEqual(recomputed_bytes, sp.total_bytes)
		self.assertEqual(prepare(dstcount).first(), stdrowcount)
		self.assertEqual(prepare(grabdst)(), stditer_tuples)

	def multiple_destinations(self, count = 3, binary = False, buffer_size = 129):
		if binary:
			src = binary_srcsql
			dst = binary_dstsql
			# accommodate for the binary header.
			count_offset = 1
		else:
			src = srcsql
			dst = dstsql
			count_offset = 0
		sqlexec(stdsource)
		dests = [new() for x in range(count)]
		receivers = []
		for x in dests:
			x.execute(stddst)
			receivers.append(copyman.StatementReceiver(x.prepare(dst)))
		sp = copyman.StatementProducer(prepare(src), buffer_size = buffer_size)
		recomputed_bytes = 0
		recomputed_messages = 0
		with copyman.CopyManager(sp, *receivers) as copy:
			for msg, bytes in copy:
				recomputed_messages += msg
				recomputed_bytes += bytes
		src_snap = prepare(grabsrc)()
		for x in dests:
			self.assertEqual(x.prepare(dstcount).first(), stdrowcount)
			self.assertEqual(x.prepare(grabdst)(), src_snap)
		self.assertEqual(stdrowcount + count_offset, recomputed_messages)
		self.assertEqual(recomputed_bytes, sp.total_bytes)

	@pg_tmp
	def testMultipleStatements(self):
		self.multiple_destinations()

	@pg_tmp
	def testMultipleStatementsBinary(self):
		self.multiple_destinations(binary = True)

	@pg_tmp
	def testMultipleStatementsSmallBuffer(self):
		self.multiple_destinations(buffer_size = 11)

	@pg_tmp
	def testNotices(self):
		# Inject a Notices directly into the stream to emulate
		# cases of asynchronous messages received during COPY.
		notices = [
			Notice((
				(b'S', b'NOTICE'),
				(b'C', b'00000'),
				(b'M', b'It\'s a beautiful day.'),
			)),
			Notice((
				(b'S', b'WARNING'),
				(b'C', b'01X1X1'),
				(b'M', b'FAILURE IS CERTAIN'),
			))
		]
		sqlexec(stdsource)
		dst = new()
		dst.execute(stddst)
		# hook for notices..
		rmessages = []
		def hook(msg):
			rmessages.append(msg)
			# suppress
			return True
		stmt = prepare(srcsql)
		stmt.msghook = hook
		sp = Injector(notices, stmt, buffer_size = 133)
		sr = copyman.StatementReceiver(dst.prepare(dstsql))
		seen_in_loop = 0
		with copyman.CopyManager(sp, sr) as copy:
			for x in copy:
				if rmessages:
					# Should get hooked before the COPY is over.
					seen_in_loop += 1
		self.assertTrue(seen_in_loop > 0)
		self.assertEqual(dst.prepare(dstcount).first(), stdrowcount)
		self.assertEqual(dst.prepare(grabdst)(), prepare(grabsrc)())
		# The injector adds then everytime the wire data is confiscated
		# from the protocol connection.
		notice, warning = rmessages[:2]
		self.assertEqual(notice.code, "00000")
		self.assertEqual(warning.code, "01X1X1")
		self.assertEqual(warning.details['severity'], "WARNING")
		self.assertEqual(notice.message, "It's a beautiful day.")
		self.assertEqual(warning.message, "FAILURE IS CERTAIN")
		self.assertEqual(notice.details['severity'], "NOTICE")

	@pg_tmp
	def testAsyncNotify(self):
		# Inject a NOTIFY directly into the stream to emulate
		# cases of asynchronous messages received during COPY.
		notify = [Notify(1234, b'channel', b'payload')]
		sqlexec(stdsource)
		dst = new()
		dst.execute(stddst)
		sp = Injector(notify, prepare(srcsql), buffer_size = 32)
		sr = copyman.StatementReceiver(dst.prepare(dstsql))
		seen_in_loop = 0
		r = []
		with copyman.CopyManager(sp, sr) as copy:
			for x in copy:
				r += list(db.iternotifies(0))
		# Got the injected NOTIFY's, right?
		self.assertTrue(r)
		# it may have happened multiple times, so adjust accordingly.
		self.assertEqual(r, [('channel', 'payload', 1234)]*len(r))

	@pg_tmp
	def testUnfinishedCopy(self):
		sqlexec(stdsource)
		dst = new()
		dst.execute(stddst)
		sp = copyman.StatementProducer(prepare(srcsql), buffer_size = 32)
		sr = copyman.StatementReceiver(dst.prepare(dstsql))
		try:
			with copyman.CopyManager(sp, sr) as copy:
				for x in copy:
					break
			self.fail("did not raise CopyFail")
		except copyman.CopyFail:
			pass

	@pg_tmp
	def testRaiseInCopy(self):
		sqlexec(stdsource)
		dst = new()
		dst.execute(stddst)
		sp = copyman.StatementProducer(prepare(srcsql), buffer_size = 128)
		sr = copyman.StatementReceiver(dst.prepare(dstsql))
		i = 0
		class ThisError(Exception):
			pass
		try:
			with copyman.CopyManager(sp, sr) as copy:
				for x in copy:
					# Note, the state of the receiver has changed.
					# We may not be on a message boundary, so this test
					# exercises cases where an interrupt occurs where
					# re-alignment *may* need to occur.
					raise ThisError()
		except copyman.CopyFail as cf:
			# It's a copy failure, but due to ThisError.
			self.assertTrue(isinstance(cf.__context__, ThisError))
		else:
			self.fail("didn't raise CopyFail")
		# Connections should be usable.
		self.assertEqual(prepare('select 1').first(), 1)
		self.assertEqual(dst.prepare('select 1').first(), 1)

	@pg_tmp
	def testRaiseInCopyOnEnter(self):
		sqlexec(stdsource)
		dst = new()
		dst.execute(stddst)
		sp = copyman.StatementProducer(prepare(srcsql), buffer_size = 128)
		sr = copyman.StatementReceiver(dst.prepare(dstsql))
		i = 0
		class ThatError(Exception):
			pass
		try:
			with copyman.CopyManager(sp, sr) as copy:
				raise ThatError()
		except copyman.CopyFail as cf:
			# yeah; error on incomplete COPY
			self.assertTrue(isinstance(cf.__context__, ThatError))
		else:
			self.fail("didn't raise CopyFail")

	@pg_tmp
	def testCopyWithFailure(self):
		sqlexec(stdsource)
		dst = new()
		dst2 = new()
		dst.execute(stddst)
		dst2.execute(stddst)
		sp = copyman.StatementProducer(prepare(srcsql), buffer_size = 128)
		sr1 = copyman.StatementReceiver(dst.prepare(dstsql))
		sr2 = copyman.StatementReceiver(dst2.prepare(dstsql))
		done = False
		with copyman.CopyManager(sp, sr1, sr2) as copy:
			while True:
				try:
					for x in copy:
						if not done:
							done = True
							dst2.pq.socket.close()
					else:
						# Done with copy.
						break
				except copyman.ReceiverFault as cf:
					if sr2 not in cf.faults:
						raise
		self.assertTrue(done)
		self.assertRaises(Exception, dst2.execute, 'select 1')
		self.assertEqual(dst.prepare(dstcount).first(), stdrowcount)
		self.assertEqual(dst.prepare(grabdst)(), prepare(grabsrc)())

	@pg_tmp
	def testEmptyRows(self):
		sqlexec(emptysource)
		dst = new()
		dst.execute(emptydst)
		sp = copyman.StatementProducer(prepare("COPY emptysource TO STDOUT"), buffer_size = 127)
		sr = copyman.StatementReceiver(dst.prepare("COPY empty FROM STDIN"))
		m = 0
		b = 0
		with copyman.CopyManager(sp, sr) as copy:
			for x in copy:
				nmsg, nbytes = x
				m += nmsg
				b += nbytes
		self.assertEqual(m, 10)
		self.assertEqual(prepare("SELECT COUNT(*) FROM emptysource").first(), 10)
		self.assertEqual(dst.prepare("SELECT COUNT(*) FROM empty").first(), 10)
		self.assertEqual(sr.count(), 10)
		self.assertEqual(sp.count(), 10)

	@pg_tmp
	def testCopyOne(self):
		from io import BytesIO
		b = BytesIO()
		copyman.transfer(
			prepare('COPY (SELECT 1) TO STDOUT'),
			copyman.CallReceiver(b.writelines)
		)
		b.seek(0)
		self.assertEqual(b.read(), b'1\n')

	@pg_tmp
	def testCopyNone(self):
		from io import BytesIO
		b = BytesIO()
		copyman.transfer(
			prepare('COPY (SELECT 1 LIMIT 0) TO STDOUT'),
			copyman.CallReceiver(b.writelines)
		)
		b.seek(0)
		self.assertEqual(b.read(), b'')

	@pg_tmp
	def testNoReceivers(self):
		sqlexec(stdsource)
		dst = new()
		dst.execute(stddst)
		sp = copyman.StatementProducer(prepare(srcsql))
		sr1 = copyman.StatementReceiver(dst.prepare(dstsql))
		done = False
		try:
			with copyman.CopyManager(sp, sr1) as copy:
				while not done:
					try:
						for x in copy:
							if not done:
								done = True
								dst.pq.socket.close()
							else:
								self.fail("failed to detect dead socket")
					except copyman.ReceiverFault as cf:
						self.assertTrue(sr1 in cf.faults)
						# Don't reconcile. Let the manager drop the receiver.
		except copyman.CopyFail:
			self.assertTrue(not bool(copy.receivers))
			# Success.
		else:
			self.fail("did not raise expected error")
		# Let the exception cause a failure.
		self.assertTrue(done)

	@pg_tmp
	def testReconciliation(self):
		# cm.reconcile() test.
		sqlexec(stdsource)
		dst = new()
		dst.execute(stddst)
		sp = copyman.StatementProducer(prepare(srcsql), buffer_size = 201)
		sr = copyman.StatementReceiver(dst.prepare(dstsql))

		original_call = sr.send
		class RecoverableError(Exception):
			pass
		def failed_write(*args):
			sr.send = original_call
			raise RecoverableError()
		sr.send = failed_write

		done = False
		recomputed_messages = 0
		recomputed_bytes = 0
		with copyman.CopyManager(sp, sr) as copy:
			while copy.receivers:
				try:
					for nmsg, nbytes in copy:
						recomputed_messages += nmsg
						recomputed_bytes += nbytes
					else:
						# Done with COPY, break out of while copy.receivers.
						break
				except copyman.ReceiverFault as cf:
					if isinstance(cf.faults[sr], RecoverableError):
						if done is True:
							self.fail("failed_write was called twice?")
						done = True
						self.assertEqual(len(copy.receivers), 0)
						copy.reconcile(sr)
						self.assertEqual(len(copy.receivers), 1)

		self.assertEqual(done, True)

		# Connections should be usable.
		self.assertEqual(prepare('select 1').first(), 1)
		self.assertEqual(dst.prepare('select 1').first(), 1)
		# validate completion
		self.assertEqual(stdrowcount, recomputed_messages)
		self.assertEqual(recomputed_bytes, sp.total_bytes)
		self.assertEqual(dst.prepare(dstcount).first(), stdrowcount)

	@pg_tmp
	def testDroppedConnection(self):
		# no cm.reconcile() test.
		sqlexec(stdsource)
		dst = new()
		dst2 = new()
		dst2.execute(stddst)
		dst.execute(stddst)
		sp = copyman.StatementProducer(prepare(srcsql), buffer_size = 201)
		sr1 = copyman.StatementReceiver(dst.prepare(dstsql))
		sr2 = copyman.StatementReceiver(dst2.prepare(dstsql))

		class TheCause(Exception):
			pass
		def failed_write(*args):
			raise TheCause()
		sr2.send = failed_write

		done = False
		recomputed_messages = 0
		recomputed_bytes = 0
		with copyman.CopyManager(sp, sr1, sr2) as copy:
			while copy.receivers:
				try:
					for nmsg, nbytes in copy:
						recomputed_messages += nmsg
						recomputed_bytes += nbytes
					else:
						# Done with COPY, break out of while copy.receivers.
						break
				except copyman.ReceiverFault as cf:
					self.assertTrue(isinstance(cf.faults[sr2], TheCause))
					if done is True:
						self.fail("failed_write was called twice?")
					done = True
					self.assertEqual(len(copy.receivers), 1)
					dst2.pq.socket.close()
					# We don't reconcile, so the manager only has one target now.

		self.assertEqual(done, True)
		# May not be aligned; really, we're expecting the connection to
		# have died.
		self.assertRaises(Exception, dst2.execute, "SELECT 1")

		# Connections should be usable.
		self.assertEqual(prepare('select 1').first(), 1)
		self.assertEqual(dst.prepare('select 1').first(), 1)
		# validate completion
		self.assertEqual(stdrowcount, recomputed_messages)
		self.assertEqual(recomputed_bytes, sp.total_bytes)
		self.assertEqual(dst.prepare(dstcount).first(), stdrowcount)
		self.assertEqual(sp.count(), stdrowcount)
		self.assertEqual(sp.command(), "COPY")

	@pg_tmp
	def testProducerFailure(self):
		sqlexec(stdsource)
		dst = new()
		dst.execute(stddst)
		sp = copyman.StatementProducer(prepare(srcsql))
		sr = copyman.StatementReceiver(dst.prepare(dstsql))
		done = False
		try:
			with copyman.CopyManager(sp, sr) as copy:
				try:
					for x in copy:
						if not done:
							done = True
							db.pq.socket.close()
				except copyman.ProducerFault as pf:
					self.assertTrue(pf.__context__ is not None)
			self.fail('expected CopyManager to raise CopyFail')
		except copyman.CopyFail as cf:
			# Expecting to see CopyFail
			self.assertTrue(True)
			self.assertTrue(isinstance(cf.producer_fault, pg_exc.ConnectionFailureError))
		self.assertTrue(done)
		self.assertRaises(Exception, sqlexec, 'select 1')
		self.assertEqual(dst.prepare(dstcount).first(), 0)

from ..copyman import WireState

class test_WireState(unittest.TestCase):
	def testNormal(self):
		WS=WireState()
		messages = WS.update(memoryview(b'd\x00\x00\x00\x04'))
		self.assertEqual(messages, 1)
		self.assertEqual(WS.remaining_bytes, 0)
		self.assertEqual(WS.size_fragment, b'')
		self.assertEqual(WS.final_view, None)

	def testIncomplete(self):
		WS=WireState()
		messages = WS.update(memoryview(b'd\x00\x00\x00\x05'))
		self.assertEqual(messages, 0)
		self.assertEqual(WS.remaining_bytes, 1)
		self.assertEqual(WS.size_fragment, b'')
		self.assertEqual(WS.final_view, None)
		messages = WS.update(memoryview(b'x'))
		self.assertEqual(messages, 1)
		self.assertEqual(WS.remaining_bytes, 0)
		self.assertEqual(WS.size_fragment, b'')
		self.assertEqual(WS.final_view, None)

	def testIncompleteHeader_0size(self):
		WS=WireState()
		messages = WS.update(memoryview(b'd'))
		self.assertEqual(messages, 0)
		self.assertEqual(WS.remaining_bytes, -1)
		self.assertEqual(WS.size_fragment, b'')
		self.assertEqual(WS.final_view, None)
		messages = WS.update(b'\x00\x00\x00\x04')
		self.assertEqual(messages, 1)

	def testIncompleteHeader_1size(self):
		WS=WireState()
		messages = WS.update(memoryview(b'd\x00'))
		self.assertEqual(messages, 0)
		self.assertEqual(WS.size_fragment, b'\x00')
		self.assertEqual(WS.final_view, None)
		self.assertEqual(WS.remaining_bytes, -1)
		messages = WS.update(memoryview(b'\x00\x00\x04'))
		self.assertEqual(messages, 1)
		self.assertEqual(WS.remaining_bytes, 0)

	def testIncompleteHeader_2size(self):
		WS=WireState()
		messages = WS.update(memoryview(b'd\x00\x00'))
		self.assertEqual(messages, 0)
		self.assertEqual(WS.remaining_bytes, -1)
		self.assertEqual(WS.size_fragment, b'\x00\x00')
		self.assertEqual(WS.final_view, None)
		messages = WS.update(b'\x00\x04')
		self.assertEqual(messages, 1)
		self.assertEqual(WS.remaining_bytes, 0)

	def testIncompleteHeader_3size(self):
		WS=WireState()
		messages = WS.update(memoryview(b'd\x00\x00\x00'))
		self.assertEqual(messages, 0)
		self.assertEqual(WS.remaining_bytes, -1)
		self.assertEqual(WS.size_fragment, b'\x00\x00\x00')
		self.assertEqual(WS.final_view, None)
		messages = WS.update(memoryview(b'\x04'))
		self.assertEqual(messages, 1)
		self.assertEqual(WS.remaining_bytes, 0)

if __name__ == '__main__':
	unittest.main()



##
# .encodings.aliases
##
"""
Module for mapping PostgreSQL encoding names to Python encoding names.

These are **not** installed in Python's aliases. Rather, `get_python_name`
should be used directly.

URLs of interest:
 * http://docs.python.org/library/codecs.html
 * http://git.postgresql.org/gitweb?p=postgresql.git;a=blob;f=src/backend/utils/mb/encnames.c
"""

##
#: Dictionary of Postgres encoding names to Python encoding names.
#: This mapping only contains those encoding names that do not intersect.
postgres_to_python = {
	'unicode' : 'utf_8',
	'sql_ascii' : 'ascii',
	'euc_jp' : 'eucjp',
	'euc_cn' : 'euccn',
	'euc_kr' : 'euckr',
	'shift_jis_2004' : 'euc_jis_2004',
	'sjis' : 'shift_jis',
	'alt' : 'cp866', # IBM866
	'abc' : 'cp1258',
	'vscii' : 'cp1258',
	'koi8r' : 'koi8_r',
	'koi8u' : 'koi8_u',
	'tcvn' : 'cp1258',
	'tcvn5712' : 'cp1258',
#	'euc_tw' : None, # N/A
#	'mule_internal' : None, # N/A
}

def get_python_name(encname):
	"""
	Lookup the name in the `postgres_to_python` dictionary. If no match is
	found, check for a 'win' or 'windows-' name and convert that to a 'cp###'
	name.

	Returns `None` if there is no alias for `encname`.

	The win[0-9]+ and windows-[0-9]+ entries are handled functionally.
	"""
	# check the dictionary first
	localname = postgres_to_python.get(encname)
	if localname is not None:
		return localname
	# no explicit mapping, check for functional transformation
	if encname.startswith('win'):
		# handle win#### and windows-####
		# remove the trailing CP number
		bare = encname.rstrip('0123456789')
		if bare.strip('_-') in ('win', 'windows'):
			return 'cp' + encname[len(bare):]
	return encname



##
# .api - ABCs for database interface elements
##
"""
Application Programmer Interfaces for PostgreSQL.

``postgresql.api`` is a collection of Python APIs for the PostgreSQL DBMS. It
is designed to take full advantage of PostgreSQL's features to provide the
Python programmer with substantial convenience.

This module is used to define "PG-API". It creates a set of ABCs
that makes up the basic interfaces used to work with a PostgreSQL server.
"""
import collections.abc
import abc

from .python.element import Element

__all__ = [
	'Message',
	'Statement',
	'Chunks',
	'Cursor',
	'Connector',
	'Category',
	'Database',
	'TypeIO',
	'Connection',
	'Transaction',
	'Settings',
	'StoredProcedure',
	'Driver',
	'Installation',
	'Cluster',
]

class Message(Element):
	"""
	A message emitted by PostgreSQL.
	A message being a NOTICE, WARNING, INFO, etc.
	"""
	_e_label = 'MESSAGE'

	severities = (
		'DEBUG',
		'INFO',
		'NOTICE',
		'WARNING',
		'ERROR',
		'FATAL',
		'PANIC',
	)
	sources = (
		'SERVER',
		'CLIENT',
	)

	@property
	@abc.abstractmethod
	def source(self) -> str:
		"""
		Where the message originated from. Normally, 'SERVER', but sometimes
		'CLIENT'.
		"""

	@property
	@abc.abstractmethod
	def code(self) -> str:
		"""
		The SQL state code of the message.
		"""

	@property
	@abc.abstractmethod
	def message(self) -> str:
		"""
		The primary message string.
		"""

	@property
	@abc.abstractmethod
	def details(self) -> dict:
		"""
		The additional details given with the message. Common keys *should* be the
		following:

		 * 'severity'
		 * 'context'
		 * 'detail'
		 * 'hint'
		 * 'file'
		 * 'line'
		 * 'function'
		 * 'position'
		 * 'internal_position'
		 * 'internal_query'
		"""

	@abc.abstractmethod
	def isconsistent(self, other) -> bool:
		"""
		Whether the fields of the `other` Message object is consistent with the
		fields of `self`.

		This *must* return the result of the comparison of code, source, message,
		and details.

		This method is provided as the alternative to overriding equality;
		often, pointer equality is the desirable means for comparison, but
		equality of the fields is also necessary.
		"""

class Result(Element):
	"""
	A result is an object managing the results of a prepared statement.

	These objects represent a binding of parameters to a given statement object.

	For results that were constructed on the server and a reference passed back
	to the client, statement and parameters may be None.
	"""
	_e_label = 'RESULT'
	_e_factors = ('statement', 'parameters', 'cursor_id')

	@abc.abstractmethod
	def close(self) -> None:
		"""
		Close the Result discarding any supporting resources and causing
		future read operations to emit empty record sets.
		"""

	@property
	@abc.abstractmethod
	def cursor_id(self) -> str:
		"""
		The cursor's identifier.
		"""

	@property
	@abc.abstractmethod
	def sql_column_types(self) -> [str]:
		"""
		The type of the columns produced by the cursor.

		A sequence of `str` objects stating the SQL type name::

			['INTEGER', 'CHARACTER VARYING', 'INTERVAL']
		"""

	@property
	@abc.abstractmethod
	def pg_column_types(self) -> [int]:
		"""
		The type Oids of the columns produced by the cursor.

		A sequence of `int` objects stating the SQL type name::

			[27, 28]
		"""

	@property
	@abc.abstractmethod
	def column_names(self) -> [str]:
		"""
		The attribute names of the columns produced by the cursor.

		A sequence of `str` objects stating the column name::

			['column1', 'column2', 'emp_name']
		"""

	@property
	@abc.abstractmethod
	def column_types(self) -> [str]:
		"""
		The Python types of the columns produced by the cursor.

		A sequence of type objects::

			[<class 'int'>, <class 'str'>]
		"""

	@property
	@abc.abstractmethod
	def parameters(self) -> (tuple, None):
		"""
		The parameters bound to the cursor. `None`, if unknown and an empty tuple
		`()`, if no parameters were given.

		These should be the *original* parameters given to the invoked statement.

		This should only be `None` when the cursor is created from an identifier,
		`postgresql.api.Database.cursor_from_id`.
		"""

	@property
	@abc.abstractmethod
	def statement(self) -> ("Statement", None):
		"""
		The query object used to create the cursor. `None`, if unknown.

		This should only be `None` when the cursor is created from an identifier,
		`postgresql.api.Database.cursor_from_id`.
		"""

@collections.abc.Iterator.register
class Chunks(Result):
	pass

@collections.abc.Iterator.register
class Cursor(Result):
	"""
	A `Cursor` object is an interface to a sequence of tuples(rows). A result
	set. Cursors publish a file-like interface for reading tuples from a cursor
	declared on the database.

	`Cursor` objects are created by invoking the `Statement.declare`
	method or by opening a cursor using an identifier via the
	`Database.cursor_from_id` method.
	"""
	_e_label = 'CURSOR'

	_seek_whence_map = {
		0 : 'ABSOLUTE',
		1 : 'RELATIVE',
		2 : 'FROM_END',
		3 : 'FORWARD',
		4 : 'BACKWARD'
	}
	_direction_map = {
		True : 'FORWARD',
		False : 'BACKWARD',
	}

	@abc.abstractmethod
	def clone(self) -> "Cursor":
		"""
		Create a new cursor using the same factors as `self`.
		"""

	def __iter__(self):
		return self

	@property
	@abc.abstractmethod
	def direction(self) -> bool:
		"""
		The default `direction` argument for read().

		When `True` reads are FORWARD.
		When `False` reads are BACKWARD.

		Cursor operation option.
		"""

	@abc.abstractmethod
	def read(self, quantity = None, direction = None) -> ["Row"]:
		"""
		Read, fetch, the specified number of rows and return them in a list.
		If quantity is `None`, all records will be fetched.

		`direction` can be used to override the default configured direction.

		This alters the cursor's position.

		Read does not directly correlate to FETCH. If zero is given as the
		quantity, an empty sequence *must* be returned.
		"""

	@abc.abstractmethod
	def __next__(self) -> "Row":
		"""
		Get the next tuple in the cursor.
		Advances the cursor position by one.
		"""

	@abc.abstractmethod
	def seek(self, offset, whence = 'ABSOLUTE'):
		"""
		Set the cursor's position to the given offset with respect to the
		whence parameter and the configured direction.

		Whence values:

		 ``0`` or ``"ABSOLUTE"``
		  Absolute.
		 ``1`` or ``"RELATIVE"``
		  Relative.
		 ``2`` or ``"FROM_END"``
		  Absolute from end.
		 ``3`` or ``"FORWARD"``
		  Relative forward.
		 ``4`` or ``"BACKWARD"``
		  Relative backward.

		Direction effects whence. If direction is BACKWARD, ABSOLUTE positioning
		will effectively be FROM_END, RELATIVE's position will be negated, and
		FROM_END will effectively be ABSOLUTE.
		"""

class Execution(metaclass = abc.ABCMeta):
	"""
	The abstract class of execution methods.
	"""

	@abc.abstractmethod
	def __call__(self, *parameters) -> ["Row"]:
		"""
		Execute the prepared statement with the given arguments as parameters.

		Usage:

		>>> p=db.prepare("SELECT column FROM ttable WHERE key = $1")
		>>> p('identifier')
		[...]
		"""

	@abc.abstractmethod
	def column(self, *parameters) -> collections.abc.Iterable:
		"""
		Return an iterator producing the values of first column of the
		rows produced by the cursor created from the statement bound with the
		given parameters.

		Column iterators are never scrollable.

		Supporting cursors will be WITH HOLD when outside of a transaction to
		allow cross-transaction access.

		`column` is designed for the situations involving large data sets.

		Each iteration returns a single value.

		column expressed in sibling terms::

			return map(operator.itemgetter(0), self.rows(*parameters))
		"""

	@abc.abstractmethod
	def chunks(self, *parameters) -> collections.abc.Iterable:
		"""
		Return an iterator producing sequences of rows produced by the cursor
		created from the statement bound with the given parameters.

		Chunking iterators are *never* scrollable.

		Supporting cursors will be WITH HOLD when outside of a transaction.

		`chunks` is designed for moving large data sets efficiently.

		Each iteration returns sequences of rows *normally* of length(seq) ==
		chunksize. If chunksize is unspecified, a default, positive integer will
		be filled in. The rows contained in the sequences are only required to
		support the basic `collections.abc.Sequence` interfaces; simple and quick
		sequence types should be used.
		"""

	@abc.abstractmethod
	def rows(self, *parameters) -> collections.abc.Iterable:
		"""
		Return an iterator producing rows produced by the cursor
		created from the statement bound with the given parameters.

		Row iterators are never scrollable.

		Supporting cursors will be WITH HOLD when outside of a transaction to
		allow cross-transaction access.

		`rows` is designed for the situations involving large data sets.

		Each iteration returns a single row. Arguably, best implemented::

			return itertools.chain.from_iterable(self.chunks(*parameters))
		"""

	@abc.abstractmethod
	def column(self, *parameters) -> collections.abc.Iterable:
		"""
		Return an iterator producing the values of the first column in
		the cursor created from the statement bound with the given parameters.

		Column iterators are never scrollable.

		Supporting cursors will be WITH HOLD when outside of a transaction to
		allow cross-transaction access.

		`column` is designed for the situations involving large data sets.

		Each iteration returns a single value. `column` is equivalent to::

			return map(operator.itemgetter(0), self.rows(*parameters))
		"""

	@abc.abstractmethod
	def declare(self, *parameters) -> Cursor:
		"""
		Return a scrollable cursor with hold using the statement bound with the
		given parameters.
		"""

	@abc.abstractmethod
	def first(self, *parameters):
		"""
		Execute the prepared statement with the given arguments as parameters.
		If the statement returns rows with multiple columns, return the first
		row. If the statement returns rows with a single column, return the
		first column in the first row. If the query does not return rows at all,
		return the count or `None` if no count exists in the completion message.

		Usage:

		>>> db.prepare("SELECT * FROM ttable WHERE key = $1").first("somekey")
		('somekey', 'somevalue')
		>>> db.prepare("SELECT 'foo'").first()
		'foo'
		>>> db.prepare("INSERT INTO atable (col) VALUES (1)").first()
		1
		"""

	@abc.abstractmethod
	def load_rows(self, iterable):
		"""
		Given an iterable, `iterable`, feed the produced parameters to the
		query. This is a bulk-loading interface for parameterized queries.

		Effectively, it is equivalent to:

			>>> q = db.prepare(sql)
			>>> for i in iterable:
			...  q(*i)

		Its purpose is to allow the implementation to take advantage of the
		knowledge that a series of parameters are to be loaded so that the
		operation can be optimized.
		"""

	@abc.abstractmethod
	def load_chunks(self, iterable):
		"""
		Given an iterable, `iterable`, feed the produced parameters of the chunks
		produced by the iterable to the query. This is a bulk-loading interface
		for parameterized queries.

		Effectively, it is equivalent to:

			>>> ps = db.prepare(...)
			>>> for c in iterable:
			...  for i in c:
			...   q(*i)

		Its purpose is to allow the implementation to take advantage of the
		knowledge that a series of chunks of parameters are to be loaded so
		that the operation can be optimized.
		"""

@collections.abc.Iterator.register
@collections.abc.Callable.register
@Execution.register
class Statement(Element):
	"""
	Instances of `Statement` are returned by the `prepare` method of
	`Database` instances.

	A Statement is an Iterable as well as Callable.

	The Iterable interface is supported for queries that take no arguments at
	all. It allows the syntax::

		>>> for x in db.prepare('select * FROM table'):
		...  pass
	"""
	_e_label = 'STATEMENT'
	_e_factors = ('database', 'statement_id', 'string',)

	@property
	@abc.abstractmethod
	def statement_id(self) -> str:
		"""
		The statment's identifier.
		"""

	@property
	@abc.abstractmethod
	def string(self) -> object:
		"""
		The SQL string of the prepared statement.

		`None` if not available. This can happen in cases where a statement is
		prepared on the server and a reference to the statement is sent to the
		client which subsequently uses the statement via the `Database`'s
		`statement` constructor.
		"""

	@property
	@abc.abstractmethod
	def sql_parameter_types(self) -> [str]:
		"""
		The type of the parameters required by the statement.

		A sequence of `str` objects stating the SQL type name::

			['INTEGER', 'VARCHAR', 'INTERVAL']
		"""

	@property
	@abc.abstractmethod
	def sql_column_types(self) -> [str]:
		"""
		The type of the columns produced by the statement.

		A sequence of `str` objects stating the SQL type name::

			['INTEGER', 'VARCHAR', 'INTERVAL']
		"""

	@property
	@abc.abstractmethod
	def pg_parameter_types(self) -> [int]:
		"""
		The type Oids of the parameters required by the statement.

		A sequence of `int` objects stating the PostgreSQL type Oid::

			[27, 28]
		"""

	@property
	@abc.abstractmethod
	def pg_column_types(self) -> [int]:
		"""
		The type Oids of the columns produced by the statement.

		A sequence of `int` objects stating the SQL type name::

			[27, 28]
		"""

	@property
	@abc.abstractmethod
	def column_names(self) -> [str]:
		"""
		The attribute names of the columns produced by the statement.

		A sequence of `str` objects stating the column name::

			['column1', 'column2', 'emp_name']
		"""

	@property
	@abc.abstractmethod
	def column_types(self) -> [type]:
		"""
		The Python types of the columns produced by the statement.

		A sequence of type objects::

			[<class 'int'>, <class 'str'>]
		"""

	@property
	@abc.abstractmethod
	def parameter_types(self) -> [type]:
		"""
		The Python types expected of parameters given to the statement.

		A sequence of type objects::

			[<class 'int'>, <class 'str'>]
		"""

	@abc.abstractmethod
	def clone(self) -> "Statement":
		"""
		Create a new statement object using the same factors as `self`.

		When used for refreshing plans, the new clone should replace references to
		the original.
		"""

	@abc.abstractmethod
	def close(self) -> None:
		"""
		Close the prepared statement releasing resources associated with it.
		"""
PreparedStatement = Statement

@collections.abc.Callable.register
class StoredProcedure(Element):
	"""
	A function stored on the database.
	"""
	_e_label = 'FUNCTION'
	_e_factors = ('database',)

	@abc.abstractmethod
	def __call__(self, *args, **kw) -> (object, Cursor, collections.abc.Iterable):
		"""
		Execute the procedure with the given arguments. If keyword arguments are
		passed they must be mapped to the argument whose name matches the key.
		If any positional arguments are given, they must fill in gaps created by
		the stated keyword arguments. If too few or too many arguments are
		given, a TypeError must be raised. If a keyword argument is passed where
		the procedure does not have a corresponding argument name, then,
		likewise, a TypeError must be raised.

		In the case where the `StoredProcedure` references a set returning
		function(SRF), the result *must* be an iterable. SRFs that return single
		columns *must* return an iterable of that column; not row data. If the
		SRF returns a composite(OUT parameters), it *should* return a `Cursor`.
		"""

##
# Arguably, it would be wiser to isolate blocks, and savepoints, but the utility
# of the separation is not significant. It's really
# more interesting as a formality that the user may explicitly state the
# type of the transaction. However, this capability is not completely absent
# from the current interface as the configuration parameters, or lack thereof,
# help imply the expectations.
class Transaction(Element):
	"""
	A `Tranaction` is an element that represents a transaction in the session.
	Once created, it's ready to be started, and subsequently committed or
	rolled back.

	Read-only transaction:

		>>> with db.xact(mode = 'read only'):
		...  ...

	Read committed isolation:

		>>> with db.xact(isolation = 'READ COMMITTED'):
		...  ...

	Savepoints are created if inside a transaction block:

		>>> with db.xact():
		...  with db.xact():
		...   ...
	"""
	_e_label = 'XACT'
	_e_factors = ('database',)

	@property
	@abc.abstractmethod
	def mode(self) -> (None, str):
		"""
		The mode of the transaction block:

			START TRANSACTION [ISOLATION] <mode>;

		The `mode` property is a string and will be directly interpolated into the
		START TRANSACTION statement.
		"""

	@property
	@abc.abstractmethod
	def isolation(self) -> (None, str):
		"""
		The isolation level of the transaction block:

			START TRANSACTION <isolation> [MODE];

		The `isolation` property is a string and will be directly interpolated into
		the START TRANSACTION statement.
		"""

	@abc.abstractmethod
	def start(self) -> None:
		"""
		Start the transaction.

		If the database is in a transaction block, the transaction should be
		configured as a savepoint. If any transaction block configuration was
		applied to the transaction, raise a `postgresql.exceptions.OperationError`.

		If the database is not in a transaction block, start one using the
		configuration where:

		`self.isolation` specifies the ``ISOLATION LEVEL``. Normally, ``READ
		COMMITTED``, ``SERIALIZABLE``, or ``READ UNCOMMITTED``.

		`self.mode` specifies the mode of the transaction. Normally, ``READ
		ONLY`` or ``READ WRITE``.

		If the transaction is already open, do nothing.

		If the transaction has been committed or aborted, raise an
		`postgresql.exceptions.OperationError`.
		"""
	begin = start

	@abc.abstractmethod
	def commit(self) -> None:
		"""
		Commit the transaction.

		If the transaction is a block, issue a COMMIT statement.

		If the transaction was started inside a transaction block, it should be
		identified as a savepoint, and the savepoint should be released.

		If the transaction has already been committed, do nothing.
		"""

	@abc.abstractmethod
	def rollback(self) -> None:
		"""
		Abort the transaction.

		If the transaction is a savepoint, ROLLBACK TO the savepoint identifier.

		If the transaction is a transaction block, issue an ABORT.

		If the transaction has already been aborted, do nothing.
		"""
	abort = rollback

	@abc.abstractmethod
	def __enter__(self):
		"""
		Run the `start` method and return self.
		"""

	@abc.abstractmethod
	def __exit__(self, typ, obj, tb):
		"""
		If an exception is indicated by the parameters, run the transaction's
		`rollback` method iff the database is still available(not closed), and
		return a `False` value.

		If an exception is not indicated, but the database's transaction state is
		in error, run the transaction's `rollback` method and raise a
		`postgresql.exceptions.InFailedTransactionError`. If the database is
		unavailable, the `rollback` method should cause a
		`postgresql.exceptions.ConnectionDoesNotExistError` exception to occur.

		Otherwise, run the transaction's `commit` method.

		When the `commit` is ultimately unsuccessful or not ran at all, the purpose
		of __exit__ is to resolve the error state of the database iff the
		database is available(not closed) so that more commands can be after the
		block's exit.
		"""

@collections.abc.MutableMapping.register
class Settings(Element):
	"""
	A mapping interface to the session's settings. This provides a direct
	interface to ``SHOW`` or ``SET`` commands. Identifiers and values need
	not be quoted specially as the implementation must do that work for the
	user.
	"""
	_e_label = 'SETTINGS'

	@abc.abstractmethod
	def __getitem__(self, key):
		"""
		Return the setting corresponding to the given key. The result should be
		consistent with what the ``SHOW`` command returns. If the key does not
		exist, raise a KeyError.
		"""

	@abc.abstractmethod
	def __setitem__(self, key, value):
		"""
		Set the setting with the given key to the given value. The action should
		be consistent with the effect of the ``SET`` command.
		"""

	@abc.abstractmethod
	def __call__(self, **kw):
		"""
		Create a context manager applying the given settings on __enter__ and
		restoring the old values on __exit__.

		>>> with db.settings(search_path = 'local,public'):
		...  ...
		"""

	@abc.abstractmethod
	def get(self, key, default = None):
		"""
		Get the setting with the corresponding key. If the setting does not
		exist, return the `default`.
		"""

	@abc.abstractmethod
	def getset(self, keys):
		"""
		Return a dictionary containing the key-value pairs of the requested
		settings. If *any* of the keys do not exist, a `KeyError` must be raised
		with the set of keys that did not exist.
		"""

	@abc.abstractmethod
	def update(self, mapping):
		"""
		For each key-value pair, incur the effect of the `__setitem__` method.
		"""

	@abc.abstractmethod
	def keys(self):
		"""
		Return an iterator to all of the settings' keys.
		"""

	@abc.abstractmethod
	def values(self):
		"""
		Return an iterator to all of the settings' values.
		"""

	@abc.abstractmethod
	def items(self):
		"""
		Return an iterator to all of the setting value pairs.
		"""

class Database(Element):
	"""
	The interface to an individual database. `Connection` objects inherit from
	this
	"""
	_e_label = 'DATABASE'

	@property
	@abc.abstractmethod
	def backend_id(self) -> (int, None):
		"""
		The backend's process identifier.
		"""

	@property
	@abc.abstractmethod
	def version_info(self) -> tuple:
		"""
		A version tuple of the database software similar Python's `sys.version_info`.

		>>> db.version_info
		(8, 1, 3, '', 0)
		"""

	@property
	@abc.abstractmethod
	def client_address(self) -> (str, None):
		"""
		The client address that the server sees. This is obtainable by querying
		the ``pg_catalog.pg_stat_activity`` relation.

		`None` if unavailable.
		"""

	@property
	@abc.abstractmethod
	def client_port(self) -> (int, None):
		"""
		The client port that the server sees. This is obtainable by querying
		the ``pg_catalog.pg_stat_activity`` relation.

		`None` if unavailable.
		"""

	@property
	@abc.abstractmethod
	def xact(self, isolation = None, mode = None) -> Transaction:
		"""
		Create a `Transaction` object using the given keyword arguments as its
		configuration.
		"""

	@property
	@abc.abstractmethod
	def settings(self) -> Settings:
		"""
		A `Settings` instance bound to the `Database`.
		"""

	@abc.abstractmethod
	def do(language, source) -> None:
		"""
		Execute a DO statement using the given language and source.
		Always returns `None`.

		Likely to be a function of Connection.execute.
		"""

	@abc.abstractmethod
	def execute(sql) -> None:
		"""
		Execute an arbitrary block of SQL. Always returns `None` and raise
		an exception on error.
		"""

	@abc.abstractmethod
	def prepare(self, sql : str) -> Statement:
		"""
		Create a new `Statement` instance bound to the connection
		using the given SQL.

		>>> s = db.prepare("SELECT 1")
		>>> c = s()
		>>> c.next()
		(1,)
		"""

	@abc.abstractmethod
	def query(self, sql : str, *args) -> Execution:
		"""
		Prepare and execute the statement, `sql`, with the given arguments.
		Equivalent to ``db.prepare(sql)(*args)``.
		"""

	@abc.abstractmethod
	def statement_from_id(self, statement_id) -> Statement:
		"""
		Create a `Statement` object that was already prepared on the
		server. The distinction between this and a regular query is that it
		must be explicitly closed if it is no longer desired, and it is
		instantiated using the statement identifier as opposed to the SQL
		statement itself.
		"""

	@abc.abstractmethod
	def cursor_from_id(self, cursor_id) -> Cursor:
		"""
		Create a `Cursor` object from the given `cursor_id` that was already
		declared on the server.

		`Cursor` objects created this way must *not* be closed when the object
		is garbage collected. Rather, the user must explicitly close it for
		the server resources to be released. This is in contrast to `Cursor`
		objects that are created by invoking a `Statement` or a SRF
		`StoredProcedure`.
		"""

	@abc.abstractmethod
	def proc(self, procedure_id) -> StoredProcedure:
		"""
		Create a `StoredProcedure` instance using the given identifier.

		The `proc_id` given can be either an ``Oid``, or a ``regprocedure``
		that identifies the stored procedure to create the interface for.

		>>> p = db.proc('version()')
		>>> p()
		'PostgreSQL 8.3.0'
		>>> qstr = "select oid from pg_proc where proname = 'generate_series'"
		>>> db.prepare(qstr).first()
		1069
		>>> generate_series = db.proc(1069)
		>>> list(generate_series(1,5))
		[1, 2, 3, 4, 5]
		"""

	@abc.abstractmethod
	def reset(self) -> None:
		"""
		Reset the connection into it's original state.

		Issues a ``RESET ALL`` to the database. If the database supports
		removing temporary tables created in the session, then remove them.
		Reapply initial configuration settings such as path.

		The purpose behind this method is to provide a soft-reconnect method
		that re-initializes the connection into its original state. One
		obvious use of this would be in a connection pool where the connection
		is being recycled.
		"""

	@abc.abstractmethod
	def notify(self, *channels, **channel_and_payload) -> int:
		"""
		NOTIFY the channels with the given payload.

		Equivalent to issuing "NOTIFY <channel>" or "NOTIFY <channel>, <payload>"
		for each item in `channels` and `channel_and_payload`. All NOTIFYs issued
		*must* occur in the same transaction.

		The items in `channels` can either be a string or a tuple. If a string,
		no payload is given, but if an item is a `builtins.tuple`, the second item
		will be given as the payload. `channels` offers a means to issue NOTIFYs
		in guaranteed order.

		The items in `channel_and_payload` are all payloaded NOTIFYs where the
		keys are the channels and the values are the payloads. Order is undefined.
		"""

	@abc.abstractmethod
	def listen(self, *channels) -> None:
		"""
		Start listening to the given channels.

		Equivalent to issuing "LISTEN <x>" for x in channels.
		"""

	@abc.abstractmethod
	def unlisten(self, *channels) -> None:
		"""
		Stop listening to the given channels.

		Equivalent to issuing "UNLISTEN <x>" for x in channels.
		"""

	@abc.abstractmethod
	def listening_channels(self) -> ["channel name", ...]:
		"""
		Return an *iterator* to all the channels currently being listened to.
		"""

	@abc.abstractmethod
	def iternotifies(self, timeout = None) -> collections.abc.Iterator:
		"""
		Return an iterator to the notifications received by the connection. The
		iterator *must* produce triples in the form ``(channel, payload, pid)``.

		If timeout is not `None`, `None` *must* be emitted at the specified
		timeout interval. If the timeout is zero, all the pending notifications
		*must* be yielded by the iterator and then `StopIteration` *must* be
		raised.

		If the connection is closed for any reason, the iterator *must* silently
		stop by raising `StopIteration`. Further error control is then the
		responsibility of the user.
		"""

class TypeIO(Element):
	_e_label = 'TYPIO'

	def _e_metas(self):
		return ()

class SocketFactory(object):
	@property
	@abc.abstractmethod
	def fatal_exception(self) -> Exception:
		"""
		The exception that is raised by sockets that indicate a fatal error.

		The exception can be a base exception as the `fatal_error_message` will
		indicate if that particular exception is actually fatal.
		"""

	@property
	@abc.abstractmethod
	def timeout_exception(self) -> Exception:
		"""
		The exception raised by the socket when an operation could not be
		completed due to a configured time constraint.
		"""

	@property
	@abc.abstractmethod
	def tryagain_exception(self) -> Exception:
		"""
		The exception raised by the socket when an operation was interrupted, but
		should be tried again.
		"""

	@property
	@abc.abstractmethod
	def tryagain(self, err : Exception) -> bool:
		"""
		Whether or not `err` suggests the operation should be tried again.
		"""

	@abc.abstractmethod
	def fatal_exception_message(self, err : Exception) -> (str, None):
		"""
		A function returning a string describing the failure, this string will be
		given to the `postgresql.exceptions.ConnectionFailure` instance that will
		subsequently be raised by the `Connection` object.

		Returns `None` when `err` is not actually fatal.
		"""

	@abc.abstractmethod
	def socket_secure(self, socket):
		"""
		Return a reference to the secured socket using the given parameters.

		If securing the socket for the connector is impossible, the user should
		never be able to instantiate the connector with parameters requesting
		security.
		"""

	@abc.abstractmethod
	def socket_factory_sequence(self) -> [collections.abc.Callable]:
		"""
		Return a sequence of `SocketCreator`s that `Connection` objects will use to
		create the socket object.
		"""

class Category(Element):
	"""
	A category is an object that initializes the subject connection for a
	specific purpose.

	Arguably, a runtime class for use with connections.
	"""
	_e_label = 'CATEGORY'
	_e_factors = ()

	@abc.abstractmethod
	def __call__(self, connection):
		"""
		Initialize the given connection in order to conform to the category.
		"""

class Connector(Element):
	"""
	A connector is an object providing the necessary information to establish a
	connection. This includes credentials, database settings, and many times
	addressing information.
	"""
	_e_label = 'CONNECTOR'
	_e_factors = ('driver', 'category')

	def __call__(self, *args, **kw):
		"""
		Create and connect. Arguments will be given to the `Connection` instance's
		`connect` method.
		"""
		return self.driver.connection(self, *args, **kw)

	def __init__(self,
		user : str = None,
		password : str = None,
		database : str = None,
		settings : (dict, [(str,str)]) = None,
		category : Category = None,
	):
		if user is None:
			# sure, it's a "required" keyword, makes for better documentation
			raise TypeError("'user' is a required keyword")
		self.user = user
		self.password = password
		self.database = database
		self.settings = settings
		self.category = category
		if category is not None and not isinstance(category, Category):
			raise TypeError("'category' must a be `None` or `postgresql.api.Category`")

class Connection(Database):
	"""
	The interface to a connection to a PostgreSQL database. This is a
	`Database` interface with the additional connection management tools that
	are particular to using a remote database.
	"""
	_e_label = 'CONNECTION'
	_e_factors = ('connector',)

	@property
	@abc.abstractmethod
	def connector(self) -> Connector:
		"""
		The :py:class:`Connector` instance facilitating the `Connection` object's
		communication and initialization.
		"""

	@property
	@abc.abstractmethod
	def closed(self) -> bool:
		"""
		`True` if the `Connection` is closed, `False` if the `Connection` is
		open.

		>>> db.closed
		True
		"""

	@abc.abstractmethod
	def clone(self) -> "Connection":
		"""
		Create another connection using the same factors as `self`. The returned
		object should be open and ready for use.
		"""

	@abc.abstractmethod
	def connect(self) -> None:
		"""
		Establish the connection to the server and initialize the category.

		Does nothing if the connection is already established.
		"""
		cat = self.connector.category
		if cat is not None:
			cat(self)

	@abc.abstractmethod
	def close(self) -> None:
		"""
		Close the connection.

		Does nothing if the connection is already closed.
		"""

	@abc.abstractmethod
	def __enter__(self):
		"""
		Establish the connection and return self.
		"""

	@abc.abstractmethod
	def __exit__(self, typ, obj, tb):
		"""
		Closes the connection and returns `False` when an exception is passed in,
		`True` when `None`.
		"""

class Driver(Element):
	"""
	The `Driver` element provides the `Connector` and other information
	pertaining to the implementation of the driver. Information about what the
	driver supports is available in instances.
	"""
	_e_label = "DRIVER"
	_e_factors = ()

	@abc.abstractmethod
	def connect(**kw):
		"""
		Create a connection using the given parameters for the Connector.
		"""

class Installation(Element):
	"""
	Interface to a PostgreSQL installation. Instances would provide various
	information about an installation of PostgreSQL accessible by the Python
	"""
	_e_label = "INSTALLATION"
	_e_factors = ()

	@property
	@abc.abstractmethod
	def version(self):
		"""
		A version string consistent with what `SELECT version()` would output.
		"""

	@property
	@abc.abstractmethod
	def version_info(self):
		"""
		A tuple specifying the version in a form similar to Python's
		sys.version_info. (8, 3, 3, 'final', 0)

		See `postgresql.versionstring`.
		"""

	@property
	@abc.abstractmethod
	def type(self):
		"""
		The "type" of PostgreSQL. Normally, the first component of the string
		returned by pg_config.
		"""

	@property
	@abc.abstractmethod
	def ssl(self) -> bool:
		"""
		Whether the installation supports SSL.
		"""

class Cluster(Element):
	"""
	Interface to a PostgreSQL cluster--a data directory. An implementation of
	this provides a means to control a server.
	"""
	_e_label = 'CLUSTER'
	_e_factors = ('installation', 'data_directory')

	@property
	@abc.abstractmethod
	def installation(self) -> Installation:
		"""
		The installation used by the cluster.
		"""

	@property
	@abc.abstractmethod
	def data_directory(self) -> str:
		"""
		The path to the data directory of the cluster.
		"""

	@abc.abstractmethod
	def init(self,
		initdb = None,
		user = None, password = None,
		encoding = None, locale = None,
		collate = None, ctype = None,
		monetary = None, numeric = None, time = None,
		text_search_config = None,
		xlogdir = None,
	):
		"""
		Create the cluster at the `data_directory` associated with the Cluster
		instance.
		"""

	@abc.abstractmethod
	def drop(self):
		"""
		Kill the server and completely remove the data directory.
		"""

	@abc.abstractmethod
	def start(self):
		"""
		Start the cluster.
		"""

	@abc.abstractmethod
	def stop(self):
		"""
		Signal the server to shutdown.
		"""

	@abc.abstractmethod
	def kill(self):
		"""
		Kill the server.
		"""

	@abc.abstractmethod
	def restart(self):
		"""
		Restart the cluster.
		"""

	@abc.abstractmethod
	def wait_until_started(self, timeout = 10):
		"""
		After the start() method is ran, the database may not be ready for use.
		This method provides a mechanism to block until the cluster is ready for
		use.

		If the `timeout` is reached, the method *must* throw a
		`postgresql.exceptions.ClusterTimeoutError`.
		"""

	@abc.abstractmethod
	def wait_until_stopped(self, timeout = 10):
		"""
		After the stop() method is ran, the database may still be running.
		This method provides a mechanism to block until the cluster is completely
		shutdown.

		If the `timeout` is reached, the method *must* throw a
		`postgresql.exceptions.ClusterTimeoutError`.
		"""

	@property
	@abc.abstractmethod
	def settings(self):
		"""
		A `Settings` interface to the ``postgresql.conf`` file associated with the
		cluster.
		"""

	@abc.abstractmethod
	def __enter__(self):
		"""
		Start the cluster if it's not already running, and wait for it to be
		readied.
		"""

	@abc.abstractmethod
	def __exit__(self, exc, val, tb):
		"""
		Stop the cluster and wait for it to shutdown *iff* it was started by the
		corresponding enter.
		"""

__docformat__ = 'reStructuredText'
if __name__ == '__main__':
	help(__package__ + '.api')



##
# .protocol.message_types
##
"""
Data module providing a sequence of bytes objects whose value corresponds to its
index in the sequence.

This provides resource for buffer objects to use common message type objects.

WARNING: It's tempting to use the 'is' operator and in some circumstances that
may be okay. However, it's possible (sys.modules.clear()) for the extension
modules' copy of this to become inconsistent with what protocol.element3 and
protocol.xact3 are using, so it's important to **not** use 'is'.
"""
message_types = tuple([bytes((x,)) for x in range(256)])



##
# .versionstring
##
"""
PostgreSQL version string parsing.

>>> postgresql.versionstring.split('8.0.1')
(8, 0, 1, None, None)
"""

def split(vstr: str) -> tuple:
	"""
	Split a PostgreSQL version string into a tuple.
	(major, minor, patch, ..., state_class, state_level)
	"""
	# bug: eg. version is 13.2 (Debian 13.2-1.pgdg100+1), so split first.
	v = vstr.strip().split()[0].split('.')

	# Get rid of the numbers around the state_class (beta,a,dev,alpha, etc)
	state_class = v[-1].strip('0123456789')
	if state_class:
		last_version, state_level = v[-1].split(state_class)
		if not state_level:
			state_level = None
		else:
			state_level = int(state_level)
		vlist = [int(x or '0') for x in v[:-1]]
		if last_version:
			vlist.append(int(last_version))
		vlist += [None] * (3 - len(vlist))
		vlist += [state_class, state_level]
	else:
		state_level = None
		state_class = None
		vlist = [int(x or '0') for x in v]
		# pad the difference with `None` objects, and +2 for the state_*.
		vlist += [None] * ((3 - len(vlist)) + 2)
	return tuple(vlist)

def unsplit(vtup: tuple) -> str:
	"""
	Join a version tuple back into the original version string.
	"""
	svtup = [str(x) for x in vtup[:-2] if x is not None]
	state_class, state_level = vtup[-2:]
	return '.'.join(svtup) + ('' if state_class is None else state_class + str(state_level))

def normalize(split_version: tuple) -> tuple:
	"""
	Given a tuple produced by `split`, normalize the `None` objects into int(0)
	or 'final' if it's the ``state_class``.
	"""
	(*head, state_class, state_level) = split_version
	mmp = [x if x is not None else 0 for x in head]
	return tuple(mmp + [state_class or 'final', state_level or 0])

default_state_class_priority = [
	'dev',
	'a',
	'alpha',
	'b',
	'beta',
	'rc',
	'final',
	None,
]

python = repr

def xml(self):
	return '<version type="one">\n' + \
		' <major>' + str(self[0]) + '</major>\n' + \
		' <minor>' + str(self[1]) + '</minor>\n' + \
		' <patch>' + str(self[2]) + '</patch>\n' + \
		' <state>' + str(self[-2]) + '</state>\n' + \
		' <level>' + str(self[-1]) + '</level>\n' + \
		'</version>'

def sh(self):
	return """PG_VERSION_MAJOR=%s
PG_VERSION_MINOR=%s
PG_VERSION_PATCH=%s
PG_VERSION_STATE=%s
PG_VERSION_LEVEL=%s""" %(
		str(self[0]),
		str(self[1]),
		str(self[2]),
		str(self[-2]),
		str(self[-1]),
	)

if __name__ == '__main__':
	import sys
	import os
	from optparse import OptionParser
	op = OptionParser()
	op.add_option('-f', '--format',
		type='choice',
		dest='format',
		help='format of output information',
		choices=('sh', 'xml', 'python'),
		default='sh',
	)
	op.add_option('-n', '--normalize',
		action='store_true',
		dest='normalize',
		help='replace missing values with defaults',
		default=False,
	)
	op.set_usage(op.get_usage().strip() + ' "version to parse"')
	co, ca = op.parse_args()
	if len(ca) != 1:
		op.error('requires exactly one argument, the version')
	else:
		v = split(ca[0])
	if co.normalize:
		v = normalize(v)
	sys.stdout.write(getattr(sys.modules[__name__], co.format)(v))
	sys.stdout.write(os.linesep)



##
# .notifyman - Receive and manage NOTIFY events.
##
"""
Notification Management Tools

Primarily this module houses the `NotificationManager` class which provides an
iterator for a NOTIFY event loop against a set of connections.

	>>> import py_opengauss
	>>> db = py_opengauss.open(...)
	>>> from py_opengauss.notifyman import NotificationManager
	>>> nm = NotificationManager(db, timeout = 10) # idle events every 10 seconds
	>>> for x in nm:
	...  if x is None:
	...   # idle event
	...   ...
	...  db, notifies = x
	...  for channel, payload, pid in notifies:
	...   ...
"""
from time import time
from select import select
from itertools import chain

class NotificationManager(object):
	"""
	A class for managing the asynchronous notifications received by a
	set of connections.

	Instances provide the iterator for an event loop that responds to NOTIFYs
	received by the connections being watched. There is no thread safety, so
	when a connection is being managed, it should not be used concurrently in
	other threads while being managed.
	"""
	__slots__ = (
		'connections',
		'garbage',
		'incoming',
		'timeout',
		'_last_time',
		'_pulled',
	)

	def __init__(self, *connections, timeout = None):
		self.settimeout(timeout)
		self.connections = set(connections)
		# Connections that failed.
		self.garbage = set()
		# Used to store NOTIFYs consumed from the connections
		self.incoming = None
		self._last_time = None
		# connection -> sequence of NOTIFYs
		self._pulled = dict()

	# Check the wire *and* wait for new messages.
	def _wait_on_wires(self, time = time, select = select):
		if self.timeout == 0:
			# We're polling.
			max_duration = 0
		else:
			# If timeout is None, we don't issue idle events, but
			# we still cycle in case the timeout is changed.
			if self._last_time is not None:
				max_duration = (self.timeout or 10) - (time() - self._last_time)
				if max_duration < 0:
					max_duration = 0
			else:
				self._last_time = time()
				max_duration = self.timeout or 10

		# Connections already marked as "bad" should not be checked.
		check = self.connections - self.garbage
		for db in check:
			if db.closed:
				self.connections.remove(db)
				self.garbage.add(db)
		check = self.connections - self.garbage

		r, w, x = select(check, (), check, max_duration)
		# Make sure the connection's _notifies get filled.
		for db in r:
			# Collect any pending events.
			try:
				# Even if db is in a failed transaction, this
				# 'null' command will succeed.
				# (only connection failures blow up)
				db.execute('')
			except Exception:
				# failed to collect notifies; put in exception list.
				# It is very unlikely that this is *not* a FATAL error.
				x.append(db)
		self.trash(x)

	def trash(self, connections):
		"""
		Remove the given connections from the set of good connections, and add
		them to the `garbage` set.

		This method can be overridden by subclasses to take a callback approach
		to connection failures.
		"""
		# Identify the bad connections.
		self.garbage.update(connections)
		self.connections.difference_update(connections)

	def queue(self, db, notifies):
		"""
		Queue the notifies for the specified connection.

		This method can be overridden by subclasses to take a callback approach
		to notification management.
		"""
		l = self._pulled.setdefault(db, list())
		l.extend(notifies)

	# Check the connection's _notifies list; just scan everything.
	def _pull_from_connections(self):
		for db in self.connections:
			if not db._notifies:
				# nothing queued up, look at the next connection
				continue
			# Pull notifies into the NotificationManager
			decode = db.typio.decode
			notifies = [
				(decode(x.channel), decode(x.payload), x.pid)
				for x in db._notifies
			]
			self.queue(db, notifies)
			del db._notifies[:len(notifies)]

	# "Append" the pulled NOTIFYs to the 'incoming' iterator.
	def _queue_next(self):
		new_seqs = []
		for db in self._pulled:
			decode = db.typio.decode
			new_seqs.append((db, self._pulled[db]))

		if new_seqs:
			if self.incoming:
				# Already have incoming; not an expected condition,
				# but let's compensate.
				self.incoming, self._pulled = chain(self.incoming, iter(new_seqs)), {}
			else:
				self.incoming, self._pulled = iter(new_seqs), {}
		elif self.incoming is None:
			# Use this to trigger the StopIteration case of zero-timeout.
			self.incoming, self._pulled = iter(()), {}

	def _timedout(self, time = time):
		# Idles are guaranteed to occur, but make sure that
		# __next__ has a chance to check the connections and the wires.
		now = time()
		if self._last_time is None:
			self._last_time = now
		elif self.timeout and now >= (self._last_time + self.timeout):
			# Set last_time to None in case the timeout is so low
			# that this condition keeps NOTIFYs from being seen.
			self._last_time = None
			# Signal timeout.
			return True
		else:
			# toggle back to None.
			self._last_time = None
		return False

	def settimeout(self, seconds):
		"""
		Set the maximum duration, in seconds, for waiting for NOTIFYs on the
		set of managed connections. The given `seconds` argument can be a number
		or `None`.

		A timeout of `None` means no timeout, and "idle" events will never
		occur.

		A timeout of `0` means to never wait for NOTIFYs. This has the effect of
		a StopIteration being raised by `__next__` when there are no more
		Notifications available for any of the connections in the set. "Idle"
		events will never occur in this situation as well.

		A timeout greater than zero means to emit `None` as "idle" events into
		the loop at the specified interval. Idle events are guaranteed to occur.
		"""
		if seconds is not None and seconds < 0:
			raise ValueError("cannot set timeout less than zero")
		self.timeout = seconds

	def gettimeout(self):
		"""
		Get the timeout assigned by `settimeout`.
		"""
		return self.timeout

	def __iter__(self):
		return self

	def __next__(self, time = time):
		checked_wire = True
		# Loop until NOTIFY received or timeout.
		while True:
			if self.incoming is not None:
				try:
					return next(self.incoming)
				except StopIteration:
					# Nothing more in this incoming.
					self.incoming = None
					# Allow a zero timeout to be used to indicate
					# that there are no NOTIFYs to be read.
					# This can be used to poll a set of
					# connections instead of listening.
					if self.timeout == 0 or not self.connections:
						raise

			# timeout happened? yield the "idle" event.
			# This check **must** happen after .incoming is checked.
			# Never emit idle when there are real events.
			if self._timedout():
				return None

			if not checked_wire and self.connections:
				# Nothing queued up, check connections if any.
				self._wait_on_wires()
				checked_wire = True
			else:
				checked_wire = False
			self._pull_from_connections()
			self._queue_next()
